@@ -4,6 +4,7 @@ set -e
 
 export NUMBA_DEVELOPER_MODE=1
 export NUMBA_DISABLE_ERROR_MESSAGE_HIGHLIGHTING=1
+export NUMBA_CAPTURED_ERRORS="new_style"
 export PYTHONFAULTHANDLER=1
 
 unamestr=`uname`
@@ -33,4 +34,5 @@ numba -s
 python -m numba.tests.test_runtests
 
 # Run the whole test suite
-$SEGVCATCH python -m numba.runtests -b -m $TEST_NPROCS
+echo "Running: $SEGVCATCH python -m numba.runtests -b -m $TEST_NPROCS -- $TESTS_TO_RUN"
+$SEGVCATCH python -m numba.runtests -b -m $TEST_NPROCS -- $TESTS_TO_RUN
