@@ -1,14 +1,15 @@
 '''
 Contains CUDA API functions
 '''
-from __future__ import absolute_import
 
+# Imports here bring together parts of the API from other modules, so some of
+# them appear unused.
 from contextlib import contextmanager
-from .cudadrv.devices import require_context, reset, gpus
+from .cudadrv.devices import require_context, reset, gpus  # noqa: F401
 from .kernel import FakeCUDAKernel
-from numba.typing import Signature
+from numba.core.typing import Signature
 from warnings import warn
-from ..args import In, Out, InOut
+from ..args import In, Out, InOut  # noqa: F401
 
 
 def select_device(dev=0):
@@ -31,6 +32,7 @@ class stream(object):
 def synchronize():
     pass
 
+
 def close():
     gpus.closed = True
 
@@ -69,24 +71,29 @@ class Event(object):
         warn('Simulator timings are bogus')
         return 0.0
 
+
 event = Event
 
 
-def jit(fn_or_sig=None, device=False, debug=False, argtypes=None, inline=False, restype=None,
-        fastmath=False, link=None):
+def jit(func_or_sig=None, device=False, debug=False, argtypes=None,
+        inline=False, restype=None, fastmath=False, link=None,
+        boundscheck=None,
+        ):
+    # Here for API compatibility
+    if boundscheck:
+        raise NotImplementedError("bounds checking is not supported for CUDA")
+
     if link is not None:
         raise NotImplementedError('Cannot link PTX in the simulator')
     # Check for first argument specifying types - in that case the
     # decorator is not being passed a function
-    if fn_or_sig is None or isinstance(fn_or_sig, (str, tuple, Signature)):
+    if func_or_sig is None or isinstance(func_or_sig, (str, tuple, Signature)):
         def jitwrapper(fn):
             return FakeCUDAKernel(fn,
                                   device=device,
                                   fastmath=fastmath)
         return jitwrapper
-    return FakeCUDAKernel(fn_or_sig, device=device)
-
-autojit = jit
+    return FakeCUDAKernel(func_or_sig, device=device)
 
 
 @contextmanager
