@@ -10,7 +10,7 @@ import traceback
 
 import numpy as np
 
-from numba import numpy_support
+from numba.np import numpy_support
 
 
 class Dim3(object):
@@ -34,6 +34,26 @@ class Dim3(object):
         yield self.z
 
 
+class GridGroup:
+    '''
+    Used to implement the grid group.
+    '''
+
+    def sync(self):
+        # Synchronization of the grid group is equivalent to synchronization of
+        # the thread block, because we only support cooperative grids with one
+        # block.
+        threading.current_thread().syncthreads()
+
+
+class FakeCUDACg:
+    '''
+    CUDA Cooperative Groups
+    '''
+    def this_grid(self):
+        return GridGroup()
+
+
 class FakeCUDALocal(object):
     '''
     CUDA Local arrays
@@ -96,38 +116,98 @@ class FakeCUDAShared(object):
             self._allocations[caller] = res
         return res
 
+
 addlock = threading.Lock()
+sublock = threading.Lock()
+andlock = threading.Lock()
+orlock = threading.Lock()
+xorlock = threading.Lock()
 maxlock = threading.Lock()
 minlock = threading.Lock()
 caslock = threading.Lock()
+inclock = threading.Lock()
+declock = threading.Lock()
+exchlock = threading.Lock()
 
 
 class FakeCUDAAtomic(object):
     def add(self, array, index, val):
         with addlock:
+            old = array[index]
             array[index] += val
+        return old
+
+    def sub(self, array, index, val):
+        with sublock:
+            old = array[index]
+            array[index] -= val
+        return old
+
+    def and_(self, array, index, val):
+        with andlock:
+            old = array[index]
+            array[index] &= val
+        return old
+
+    def or_(self, array, index, val):
+        with orlock:
+            old = array[index]
+            array[index] |= val
+        return old
+
+    def xor(self, array, index, val):
+        with xorlock:
+            old = array[index]
+            array[index] ^= val
+        return old
+
+    def inc(self, array, index, val):
+        with inclock:
+            old = array[index]
+            if old >= val:
+                array[index] = 0
+            else:
+                array[index] += 1
+        return old
+
+    def dec(self, array, index, val):
+        with declock:
+            old = array[index]
+            if (old == 0) or (old > val):
+                array[index] = val
+            else:
+                array[index] -= 1
+        return old
+
+    def exch(self, array, index, val):
+        with exchlock:
+            old = array[index]
+            array[index] = val
+        return old
 
     def max(self, array, index, val):
         with maxlock:
-            # CUDA Python's semantics for max differ from Numpy's Python's,
-            # so we have special handling here (CUDA Python treats NaN as
-            # missing data).
-            if np.isnan(array[index]):
-                array[index] = val
-            elif np.isnan(val):
-                return
-            array[index] = max(array[index], val)
+            old = array[index]
+            array[index] = max(old, val)
+        return old
 
     def min(self, array, index, val):
         with minlock:
-            # CUDA Python's semantics for min differ from Numpy's Python's,
-            # so we have special handling here (CUDA Python treats NaN as
-            # missing data).
-            if np.isnan(array[index]):
-                array[index] = val
-            elif np.isnan(val):
-                return
-            array[index] = min(array[index], val)
+            old = array[index]
+            array[index] = min(old, val)
+        return old
+
+    def nanmax(self, array, index, val):
+        with maxlock:
+            old = array[index]
+            array[index] = np.nanmax([array[index], val])
+        return old
+
+    def nanmin(self, array, index, val):
+        with minlock:
+            old = array[index]
+            array[index] = np.nanmin([array[index], val])
+        return old
 
     def compare_and_swap(self, array, old, val):
         with caslock:
@@ -138,6 +218,26 @@ class FakeCUDAAtomic(object):
             return loaded
 
 
+class FakeCUDAFp16(object):
+    def hadd(self, a, b):
+        return a + b
+
+    def hsub(self, a, b):
+        return a - b
+
+    def hmul(self, a, b):
+        return a * b
+
+    def hfma(self, a, b, c):
+        return a * b + c
+
+    def hneg(self, a):
+        return -a
+
+    def habs(self, a):
+        return abs(a)
+
+
 class FakeCUDAModule(object):
     '''
     An instance of this class will be injected into the __globals__ for an
@@ -152,10 +252,16 @@ class FakeCUDAModule(object):
     def __init__(self, grid_dim, block_dim, dynshared_size):
         self.gridDim = Dim3(*grid_dim)
         self.blockDim = Dim3(*block_dim)
+        self._cg = FakeCUDACg()
         self._local = FakeCUDALocal()
         self._shared = FakeCUDAShared(dynshared_size)
         self._const = FakeCUDAConst()
         self._atomic = FakeCUDAAtomic()
+        self._fp16 = FakeCUDAFp16()
+
+    @property
+    def cg(self):
+        return self._cg
 
     @property
     def local(self):
@@ -173,6 +279,10 @@ class FakeCUDAModule(object):
     def atomic(self):
         return self._atomic
 
+    @property
+    def fp16(self):
+        return self._fp16
+
     @property
     def threadIdx(self):
         return threading.current_thread().threadIdx
@@ -216,6 +326,12 @@ class FakeCUDAModule(object):
     def popc(self, val):
         return bin(val).count("1")
 
+    def fma(self, a, b, c):
+        return a * b + c
+
+    def cbrt(self, a):
+        return a ** (1 / 3)
+
     def brev(self, val):
         return int('{:032b}'.format(val)[::-1], 2)
 
@@ -224,8 +340,14 @@ class FakeCUDAModule(object):
         return len(s) - len(s.lstrip('0'))
 
     def ffs(self, val):
+        # The algorithm is:
+        # 1. Count the number of trailing zeros.
+        # 2. Add 1, because the LSB is numbered 1 rather than 0, and so on.
+        # 3. If we've counted 32 zeros (resulting in 33), there were no bits
+        #    set so we need to return zero.
         s = '{:032b}'.format(val)
-        return len(s) - len(s.rstrip('0'))
+        r = (len(s) - len(s.rstrip('0')) + 1) % 33
+        return r
 
     def selp(self, a, b, c):
         return b if a else c
