@@ -1,21 +1,20 @@
 # -*- coding: utf-8 -*-
-from __future__ import print_function, division, absolute_import
 
 import logging
 import os
 import sys
 
 from llvmlite import ir
+from llvmlite.binding import Linkage
 import llvmlite.llvmpy.core as lc
 
-from numba import cgutils
-from numba.utils import IS_PY3
-from . import llvm_types as lt
-from numba.compiler import compile_extra, Flags
-from numba.compiler_lock import global_compiler_lock
+from numba.pycc import llvm_types as lt
+from numba.core.compiler import compile_extra, Flags
+from numba.core.compiler_lock import global_compiler_lock
 
-from numba.targets.registry import cpu_target
-from numba.runtime import nrtdynmod
+from numba.core.registry import cpu_target
+from numba.core.runtime import nrtdynmod
+from numba.core import cgutils
 
 
 logger = logging.getLogger(__name__)
@@ -100,8 +99,8 @@ class _ModuleCompiler(object):
                                     lt._int8_star))
 
     method_def_ptr = lc.Type.pointer(method_def_ty)
-
-    env_def_ty = lc.Type.struct((lt._void_star, lt._int32))
+    # The structure type constructed by PythonAPI.serialize_uncached()
+    env_def_ty = lc.Type.struct((lt._void_star, lt._int32, lt._void_star))
     env_def_ptr = lc.Type.pointer(env_def_ty)
 
     def __init__(self, export_entries, module_name, use_nrt=False,
@@ -140,11 +139,12 @@ class _ModuleCompiler(object):
 
         # Generate IR for all exported functions
         flags = Flags()
-        flags.set("no_compile")
+        flags.no_compile = True
         if not self.export_python_wrap:
-            flags.set("no_cpython_wrapper")
+            flags.no_cpython_wrapper = True
+            flags.no_cfunc_wrapper = True
         if self.use_nrt:
-            flags.set("nrt")
+            flags.nrt = True
             # Compile NRT helpers
             nrt_module, _ = nrtdynmod.create_nrt_module(self.context)
             library.add_ir_module(nrt_module)
@@ -183,8 +183,11 @@ class _ModuleCompiler(object):
         library.finalize()
         for fn in library.get_defined_functions():
             if fn.name not in self.dll_exports:
-                fn.visibility = "hidden"
-
+                if fn.linkage in {Linkage.private, Linkage.internal}:
+                    # Private/Internal linkage must have "default" visibility
+                    fn.visibility = "default"
+                else:
+                    fn.visibility = 'hidden'
         return library
 
     def write_llvm_bitcode(self, output, wrap=False, **kws):
@@ -229,7 +232,7 @@ class _ModuleCompiler(object):
             name = entry.symbol
             llvm_func_name = self._mangle_method_symbol(name)
             fnty = self.exported_function_types[entry]
-            lfunc = llvm_module.add_function(fnty, name=llvm_func_name)
+            lfunc = ir.Function(llvm_module, fnty, llvm_func_name)
 
             method_name = self.context.insert_const_string(llvm_module, name)
             method_def_const = lc.Constant.struct((method_name,
@@ -241,8 +244,9 @@ class _ModuleCompiler(object):
         sentinel = lc.Constant.struct([NULL, NULL, ZERO, NULL])
         method_defs.append(sentinel)
         method_array_init = lc.Constant.array(self.method_def_ty, method_defs)
-        method_array = llvm_module.add_global_variable(method_array_init.type,
-                                                       '.module_methods')
+        method_array = cgutils.add_global_variable(llvm_module,
+                                                   method_array_init.type,
+                                                   '.module_methods')
         method_array.initializer = method_array_init
         method_array.linkage = lc.LINKAGE_INTERNAL
         method_array_ptr = lc.Constant.gep(method_array, [ZERO, ZERO])
@@ -293,79 +297,14 @@ class _ModuleCompiler(object):
             fnty = ir.FunctionType(lt._int32,
                                    [modobj.type, self.method_def_ptr,
                                     self.env_def_ptr, envgv_array.type])
-            fn = llvm_module.add_function(fnty, self.external_init_function)
+            fn = ir.Function(llvm_module, fnty, self.external_init_function)
             return builder.call(fn, [modobj, method_array, env_array,
                                      envgv_array])
         else:
             return None
 
 
-class ModuleCompilerPy2(_ModuleCompiler):
-
-    @property
-    def module_create_definition(self):
-        """Return the signature and name of the function to initialize the module.
-        """
-        signature = lc.Type.function(lt._pyobject_head_p,
-                                     (lt._int8_star,
-                                      self.method_def_ptr,
-                                      lt._int8_star,
-                                      lt._pyobject_head_p,
-                                      lt._int32))
-
-        name = "Py_InitModule4"
-
-        if lt._trace_refs_:
-            name += "TraceRefs"
-        if lt._plat_bits == 64:
-            name += "_64"
-
-        return signature, name
-
-    @property
-    def module_init_definition(self):
-        """Return the signature and name of the function to initialize the extension.
-        """
-        return lc.Type.function(lc.Type.void(), ()), "init" + self.module_name
-
-    def _emit_python_wrapper(self, llvm_module):
-
-        # Define the module initialization function.
-        mod_init_fn = llvm_module.add_function(*self.module_init_definition)
-        entry = mod_init_fn.append_basic_block('Entry')
-        builder = lc.Builder(entry)
-        pyapi = self.context.get_python_api(builder)
-
-        # Python C API module creation function.
-        create_module_fn = llvm_module.add_function(*self.module_create_definition)
-        create_module_fn.linkage = lc.LINKAGE_EXTERNAL
-
-        # Define a constant string for the module name.
-        mod_name_const = self.context.insert_const_string(llvm_module,
-                                                          self.module_name)
-
-        method_array = self._emit_method_array(llvm_module)
-
-        mod = builder.call(create_module_fn,
-                           (mod_name_const,
-                            method_array,
-                            NULL,
-                            lc.Constant.null(lt._pyobject_head_p),
-                            lc.Constant.int(lt._int32, sys.api_version)))
-
-        env_array = self._emit_environment_array(llvm_module, builder, pyapi)
-        envgv_array = self._emit_envgvs_array(llvm_module, builder, pyapi)
-
-        self._emit_module_init_code(llvm_module, builder, mod,
-                                    method_array, env_array, envgv_array)
-        # XXX No way to notify failure to caller...
-
-        builder.ret_void()
-
-        self.dll_exports.append(mod_init_fn.name)
-
-
-class ModuleCompilerPy3(_ModuleCompiler):
+class ModuleCompiler(_ModuleCompiler):
 
     _ptr_fun = lambda ret, *args: lc.Type.pointer(lc.Type.function(ret, args))
 
@@ -453,7 +392,7 @@ class ModuleCompilerPy3(_ModuleCompiler):
     def _emit_python_wrapper(self, llvm_module):
         # Figure out the Python C API module creation function, and
         # get a LLVM function for it.
-        create_module_fn = llvm_module.add_function(*self.module_create_definition)
+        create_module_fn = ir.Function(llvm_module, *self.module_create_definition)
         create_module_fn.linkage = lc.LINKAGE_EXTERNAL
 
         # Define a constant string for the module name.
@@ -467,8 +406,9 @@ class ModuleCompilerPy3(_ModuleCompiler):
              lc.Constant.null(lt._pyobject_head_p),         # m_copy
             )
         )
-        mod_def_base = llvm_module.add_global_variable(mod_def_base_init.type,
-                                                       '.module_def_base')
+        mod_def_base = cgutils.add_global_variable(llvm_module,
+                                                   mod_def_base_init.type,
+                                                   '.module_def_base')
         mod_def_base.initializer = mod_def_base_init
         mod_def_base.linkage = lc.LINKAGE_INTERNAL
 
@@ -488,13 +428,13 @@ class ModuleCompilerPy3(_ModuleCompiler):
         )
 
         # Define a constant string for the module name.
-        mod_def = llvm_module.add_global_variable(mod_def_init.type,
-                                                  '.module_def')
+        mod_def = cgutils.add_global_variable(llvm_module, mod_def_init.type,
+                                              '.module_def')
         mod_def.initializer = mod_def_init
         mod_def.linkage = lc.LINKAGE_INTERNAL
 
         # Define the module initialization function.
-        mod_init_fn = llvm_module.add_function(*self.module_init_definition)
+        mod_init_fn = ir.Function(llvm_module, *self.module_init_definition)
         entry = mod_init_fn.append_basic_block('Entry')
         builder = lc.Builder(entry)
         pyapi = self.context.get_python_api(builder)
@@ -522,5 +462,3 @@ class ModuleCompilerPy3(_ModuleCompiler):
 
         self.dll_exports.append(mod_init_fn.name)
 
-
-ModuleCompiler = ModuleCompilerPy3 if IS_PY3 else ModuleCompilerPy2
