@@ -1,23 +1,21 @@
-from __future__ import print_function, division, absolute_import
-
-import numba.unittest_support as unittest
-
 import collections
 import contextlib
 import cProfile
+import inspect
 import gc
 import multiprocessing
 import os
 import random
 import sys
 import time
+import unittest
 import warnings
 
+from io import StringIO
 from unittest import result, runner, signals, suite, loader, case
 
 from .loader import TestLoader
-from numba.utils import PYVERSION, StringIO
-from numba import config
+from numba.core import config
 
 try:
     from multiprocessing import TimeoutError
@@ -55,17 +53,51 @@ def make_tag_decorator(known_tags):
     return tag
 
 
+def cuda_sensitive_mtime(x):
+    """
+    Return a key for sorting tests bases on mtime and test name. For CUDA
+    tests, interleaving tests from different classes is dangerous as the CUDA
+    context might get reset unexpectedly between methods of a class, so for
+    CUDA tests the key prioritises the test module and class ahead of the
+    mtime.
+    """
+    cls = x.__class__
+    key = str(os.path.getmtime(inspect.getfile(cls))) + str(x)
+
+    from numba.cuda.testing import CUDATestCase
+    if CUDATestCase in cls.mro():
+        key = "%s.%s %s" % (str(cls.__module__), str(cls.__name__), key)
+
+    return key
+
+def parse_slice(useslice):
+    """Parses the argument string "useslice" as the arguments to the `slice()`
+    constructor and returns a slice object that's been instantiated with those
+    arguments. i.e. input useslice="1,20,2" leads to output `slice(1, 20, 2)`.
+    """
+    try:
+        l = {}
+        exec("sl = slice(%s)" % useslice, l)
+        return l['sl']
+    except Exception:
+        msg = ("Expected arguments consumable by 'slice' to follow "
+                "option `-j`, found '%s'" % useslice)
+        raise ValueError(msg)
+
+
 class TestLister(object):
     """Simply list available tests rather than running them."""
-    def __init__(self):
-        pass
+    def __init__(self, useslice):
+        self.useslice = parse_slice(useslice)
 
     def run(self, test):
         result = runner.TextTestResult(sys.stderr, descriptions=True, verbosity=1)
         self._test_list = _flatten_suite(test)
-        for t in self._test_list:
+        masked_list = self._test_list[self.useslice]
+        self._test_list.sort(key=cuda_sensitive_mtime)
+        for t in masked_list:
             print(t.id())
-        print('%d tests found' % len(self._test_list))
+        print('%d tests found. %s selected' % (len(self._test_list), len(masked_list)))
         return result
 
 
@@ -89,6 +121,18 @@ class SerialSuite(unittest.TestSuite):
             super(SerialSuite, self).addTest(test)
 
 
+class BasicTestRunner(runner.TextTestRunner):
+    def __init__(self, useslice, **kwargs):
+        runner.TextTestRunner.__init__(self, **kwargs)
+        self.useslice = parse_slice(useslice)
+
+    def run(self, test):
+        run = _flatten_suite(test)[self.useslice]
+        run.sort(key=cuda_sensitive_mtime)
+        wrapped = unittest.TestSuite(run)
+        return super(BasicTestRunner, self).run(wrapped)
+
+
 # "unittest.main" is really the TestProgram class!
 # (defined in a module named itself "unittest.main"...)
 
@@ -106,17 +150,14 @@ class NumbaTestProgram(unittest.main):
     refleak = False
     profile = False
     multiprocess = False
+    useslice = None
     list = False
     tags = None
+    exclude_tags = None
     random_select = None
     random_seed = 42
 
     def __init__(self, *args, **kwargs):
-        # Disable interpreter fallback if we are running the test suite
-        if config.COMPATIBILITY_MODE:
-            warnings.warn("Unset INTERPRETER_FALLBACK")
-            config.COMPATIBILITY_MODE = False
-
         topleveldir = kwargs.pop('topleveldir', None)
         kwargs['testLoader'] = TestLoader(topleveldir)
 
@@ -146,44 +187,72 @@ class NumbaTestProgram(unittest.main):
         parser.add_argument('--tags', dest='tags', type=str,
                             help='Comma-separated list of tags to select '
                                  'a subset of the test suite')
+        parser.add_argument('--exclude-tags', dest='exclude_tags', type=str,
+                            help='Comma-separated list of tags to de-select '
+                                 'a subset of the test suite')
         parser.add_argument('--random', dest='random_select', type=float,
                             help='Random proportion of tests to select')
         parser.add_argument('--profile', dest='profile',
                             action='store_true',
                             help='Profile the test run')
+        parser.add_argument('-j', '--slice', dest='useslice', nargs='?',
+                            type=str, const="None",
+                            help='Slice the test sequence')
+
+        def git_diff_str(x):
+            if x != 'ancestor':
+                raise ValueError("invalid option for --gitdiff")
+            return x
+
+        parser.add_argument('-g', '--gitdiff', dest='gitdiff', type=git_diff_str,
+                            default=False, nargs='?',
+                            help=('Run tests from changes made against '
+                                  'origin/master as identified by `git diff`. '
+                                  'If set to "ancestor", the diff compares '
+                                  'against the common ancestor.'))
         return parser
 
+    def _handle_tags(self, argv, tagstr):
+        found = None
+        for x in argv:
+            if tagstr in x:
+                if found is None:
+                    found = x
+                else:
+                    raise ValueError("argument %s supplied repeatedly" % tagstr)
+
+        if found is not None:
+            posn = argv.index(found)
+            try:
+                if found == tagstr: # --tagstr <arg>
+                    tag_args = argv[posn + 1].strip()
+                    argv.remove(tag_args)
+                else: # --tagstr=<arg>
+                    if '=' in found:
+                        tag_args =  found.split('=')[1].strip()
+                    else:
+                        raise AssertionError('unreachable')
+            except IndexError:
+                # at end of arg list, raise
+                msg = "%s requires at least one tag to be specified"
+                raise ValueError(msg % tagstr)
+            # see if next arg is "end options" or some other flag
+            if tag_args.startswith('-'):
+                raise ValueError("tag starts with '-', probably a syntax error")
+            # see if tag is something like "=<tagname>" which is likely a syntax
+            # error of form `--tags =<tagname>`, note the space prior to `=`.
+            if '=' in tag_args:
+                msg = "%s argument contains '=', probably a syntax error"
+                raise ValueError(msg % tagstr)
+            attr = tagstr[2:].replace('-', '_')
+            setattr(self, attr, tag_args)
+            argv.remove(found)
+
+
     def parseArgs(self, argv):
         if '-l' in argv:
             argv.remove('-l')
             self.list = True
-        if PYVERSION < (3, 4) and '-m' in argv:
-            # We want '-m' to work on all versions, emulate this option.
-            dashm_posn = argv.index('-m')
-            # the default number of processes to use
-            nprocs = multiprocessing.cpu_count()
-            # see what else is in argv
-            # ensure next position is safe for access
-            try:
-                m_option = argv[dashm_posn + 1]
-                # see if next arg is "end options"
-                if m_option != '--':
-                    #try and parse the next arg as an int
-                    try:
-                        nprocs = int(m_option)
-                    except BaseException:
-                        msg = ('Expected an integer argument to '
-                               'option `-m`, found "%s"')
-                        raise ValueError(msg % m_option)
-                    # remove the value of the option
-                    argv.remove(m_option)
-                # else end options, use defaults
-            except IndexError:
-                # at end of arg list, use defaults
-                pass
-
-            self.multiprocess = nprocs
-            argv.remove('-m')
 
         super(NumbaTestProgram, self).parseArgs(argv)
 
@@ -195,12 +264,22 @@ class NumbaTestProgram(unittest.main):
 
         if self.tags:
             tags = [s.strip() for s in self.tags.split(',')]
-            self.test = _choose_tagged_tests(self.test, tags)
+            self.test = _choose_tagged_tests(self.test, tags, mode='include')
+
+        if self.exclude_tags:
+            tags = [s.strip() for s in self.exclude_tags.split(',')]
+            self.test = _choose_tagged_tests(self.test, tags, mode='exclude')
 
         if self.random_select:
             self.test = _choose_random_tests(self.test, self.random_select,
                                              self.random_seed)
 
+        if self.gitdiff is not False:
+            self.test = _choose_gitdiff_tests(
+                self.test,
+                use_common_ancestor=(self.gitdiff == 'ancestor'),
+            )
+
         if self.verbosity <= 0:
             # We aren't interested in informational messages / warnings when
             # running with '-q'.
@@ -222,18 +301,22 @@ class NumbaTestProgram(unittest.main):
                               "of Python, only memory leaks will be detected")
 
         elif self.list:
-            self.testRunner = TestLister()
+            self.testRunner = TestLister(self.useslice)
 
         elif self.testRunner is None:
-            self.testRunner = unittest.TextTestRunner
+            self.testRunner = BasicTestRunner(self.useslice,
+                                              verbosity=self.verbosity,
+                                              failfast=self.failfast,
+                                              buffer=self.buffer)
 
         if self.multiprocess and not self.nomultiproc:
             if self.multiprocess < 1:
                 msg = ("Value specified for the number of processes to use in "
                     "running the suite must be > 0")
                 raise ValueError(msg)
-            self.testRunner = ParallelTestRunner(self.testRunner,
+            self.testRunner = ParallelTestRunner(runner.TextTestRunner,
                                                  self.multiprocess,
+                                                 self.useslice,
                                                  verbosity=self.verbosity,
                                                  failfast=self.failfast,
                                                  buffer=self.buffer)
@@ -257,22 +340,86 @@ class NumbaTestProgram(unittest.main):
             run_tests_real()
 
 
+# These are tests which are generated and injected into the test suite, what
+# gets injected depends on features of the test environment, e.g. TBB presence
+# it's important for doing the CI "slice tests" that these are run at the end
+# See notes in `_flatten_suite` for why. Simple substring matching is used to
+# determine a match.
+_GENERATED = (
+    "numba.cuda.tests.cudapy.test_libdevice.TestLibdeviceCompilation",
+    "numba.tests.test_num_threads",
+    "numba.tests.test_parallel_backend",
+    "numba.tests.test_svml",
+    "numba.tests.test_ufuncs",
+)
+
+
+def _flatten_suite_inner(test):
+    """
+    Workhorse for _flatten_suite
+    """
+    tests = []
+    if isinstance(test, (unittest.TestSuite, list, tuple)):
+        for x in test:
+            tests.extend(_flatten_suite_inner(x))
+    else:
+        tests.append(test)
+    return tests
+
+
 def _flatten_suite(test):
     """
     Expand nested suite into list of test cases.
     """
-    if isinstance(test, (unittest.TestSuite, list, tuple)):
-        tests = []
-        for x in test:
-            tests.extend(_flatten_suite(x))
-        return tests
+    tests = _flatten_suite_inner(test)
+    # Strip out generated tests and stick them at the end, this is to make sure
+    # that tests appear in a consistent order regardless of features available.
+    # This is so that a slice through the test suite e.g. (1::N) would likely be
+    # consistent up to the point of the generated tests, which rely on specific
+    # features.
+    generated = set()
+    for t in tests:
+        for g in _GENERATED:
+            if g in str(t):
+                generated.add(t)
+    normal = set(tests) - generated
+    def key(x):
+        return x.__module__, type(x).__name__, x._testMethodName
+    tests = sorted(normal, key=key)
+    tests.extend(sorted(list(generated), key=key))
+    return tests
+
+
+def _choose_gitdiff_tests(tests, *, use_common_ancestor=False):
+    try:
+        from git import Repo
+    except ImportError:
+        raise ValueError("gitpython needed for git functionality")
+    repo = Repo('.')
+    path = os.path.join('numba', 'tests')
+    if use_common_ancestor:
+        print(f"Git diff by common ancestor")
+        target = 'origin/master...HEAD'
     else:
-        return [test]
+        target = 'origin/master..HEAD'
+    gdiff_paths = repo.git.diff(target, path, name_only=True).split()
+    # normalise the paths as they are unix style from repo.git.diff
+    gdiff_paths = [os.path.normpath(x) for x in gdiff_paths]
+    selected = []
+    gdiff_paths = [os.path.join(repo.working_dir, x) for x in gdiff_paths]
+    for test in _flatten_suite(tests):
+        assert isinstance(test, unittest.TestCase)
+        fname = inspect.getsourcefile(test.__class__)
+        if fname in gdiff_paths:
+            selected.append(test)
+    print("Git diff identified %s tests" % len(selected))
+    return unittest.TestSuite(selected)
 
-
-def _choose_tagged_tests(tests, tags):
+def _choose_tagged_tests(tests, tags, mode='include'):
     """
-    Select tests that are tagged with at least one of the given tags.
+    Select tests that are tagged/not tagged with at least one of the given tags.
+    Set mode to 'include' to include the tests with tags, or 'exclude' to
+    exclude the tests with the tags.
     """
     selected = []
     tags = set(tags)
@@ -284,12 +431,18 @@ def _choose_tagged_tests(tests, tags):
             func = func.im_func
         except AttributeError:
             pass
-        try:
-            if func.tags & tags:
+
+        found_tags = getattr(func, 'tags', None)
+        # only include the test if the tags *are* present
+        if mode == 'include':
+            if found_tags is not None and found_tags & tags:
                 selected.append(test)
-        except AttributeError:
-            # Test method doesn't have any tags
-            pass
+        elif mode == 'exclude':
+            # only include the test if the tags *are not* present
+            if found_tags is None or not (found_tags & tags):
+                selected.append(test)
+        else:
+            raise ValueError("Invalid 'mode' supplied: %s." % mode)
     return unittest.TestSuite(selected)
 
 
@@ -528,13 +681,15 @@ class _MinimalRunner(object):
                 del test.__dict__[name]
 
 
-def _split_nonparallel_tests(test):
+def _split_nonparallel_tests(test, sliced=slice(None)):
     """
     Split test suite into parallel and serial tests.
     """
     ptests = []
     stests = []
 
+    flat = _flatten_suite(test)[sliced]
+
     def is_parallelizable_test_case(test):
         # Guard for the fake test case created by unittest when test
         # discovery fails, as it isn't picklable (e.g. "LoadTestsFailure")
@@ -545,20 +700,16 @@ def _split_nonparallel_tests(test):
         # Was parallel execution explicitly disabled?
         return getattr(test, "_numba_parallel_test_", True)
 
-    if isinstance(test, unittest.TestSuite):
-        # It's a sub-suite, recurse
-        for t in test:
-            p, s = _split_nonparallel_tests(t)
-            ptests.extend(p)
-            stests.extend(s)
-    elif is_parallelizable_test_case(test):
-        # Test case is suitable for parallel execution (default)
-        ptests = [test]
-    else:
-        # Test case explicitly disallows parallel execution
-        stests = _flatten_suite(test)
+    for t in flat:
+        if is_parallelizable_test_case(t):
+            ptests.append(t)
+        else:
+            stests.append(t)
+
     return ptests, stests
 
+# A test can't run longer than 10 minutes
+_TIMEOUT = 600
 
 class ParallelTestRunner(runner.TextTestRunner):
     """
@@ -567,13 +718,13 @@ class ParallelTestRunner(runner.TextTestRunner):
     """
 
     resultclass = ParallelTestResult
-    # A test can't run longer than 5 minutes
-    timeout = 300
+    timeout = _TIMEOUT
 
-    def __init__(self, runner_cls, nprocs, **kwargs):
+    def __init__(self, runner_cls, nprocs, useslice, **kwargs):
         runner.TextTestRunner.__init__(self, **kwargs)
         self.runner_cls = runner_cls
         self.nprocs = nprocs
+        self.useslice = parse_slice(useslice)
         self.runner_args = kwargs
 
     def _run_inner(self, result):
@@ -582,7 +733,7 @@ class ParallelTestRunner(runner.TextTestRunner):
         child_runner = _MinimalRunner(self.runner_cls, self.runner_args)
 
         # Split the tests and recycle the worker process to tame memory usage.
-        chunk_size = 500
+        chunk_size = 100
         splitted_tests = [self._ptests[i:i + chunk_size]
                           for i in range(0, len(self._ptests), chunk_size)]
 
@@ -612,6 +763,7 @@ class ParallelTestRunner(runner.TextTestRunner):
 
     def _run_parallel_tests(self, result, pool, child_runner, tests):
         remaining_ids = set(t.id() for t in tests)
+        tests.sort(key=cuda_sensitive_mtime)
         it = pool.imap_unordered(child_runner, tests)
         while True:
             try:
@@ -633,7 +785,11 @@ class ParallelTestRunner(runner.TextTestRunner):
                     return
 
     def run(self, test):
-        self._ptests, self._stests = _split_nonparallel_tests(test)
+        self._ptests, self._stests = _split_nonparallel_tests(test,
+                                                              sliced=
+                                                              self.useslice)
+        print("Parallel: %s. Serial: %s" % (len(self._ptests),
+                                            len(self._stests)))
         # This will call self._run_inner() on the created result object,
         # and print out the detailed test results at the end.
         return super(ParallelTestRunner, self).run(self._run_inner)
