@@ -1,33 +1,60 @@
-from __future__ import division
-
 import itertools
 
 import numpy as np
 import sys
 from collections import namedtuple
+from io import StringIO
 
-from numba import unittest_support as unittest
-from numba import njit, typeof, types, typing, typeof, ir, utils, bytecode
-from .support import TestCase, tag
-from numba.array_analysis import EquivSet, ArrayAnalysis
-from numba.compiler import Pipeline, Flags, _PipelineManager
-from numba.targets import cpu, registry
-from numba.numpy_support import version as numpy_version
-from numba.ir_utils import remove_dead
+from numba import njit, typeof, prange
+from numba.core import (
+    types,
+    typing,
+    ir,
+    bytecode,
+    postproc,
+    cpu,
+    registry,
+    utils,
+)
+from numba.tests.support import (TestCase, tag, skip_parfors_unsupported,
+                                 skip_unless_scipy)
+from numba.parfors.array_analysis import EquivSet, ArrayAnalysis
+from numba.core.compiler import Compiler, Flags, PassManager
+from numba.core.ir_utils import remove_dead
+from numba.core.untyped_passes import (ExtractByteCode, TranslateByteCode, FixupArgs,
+                             IRProcessing, DeadBranchPrune,
+                             RewriteSemanticConstants, GenericRewrites,
+                             WithLifting, PreserveIR, InlineClosureLikes)
+
+from numba.core.typed_passes import (NopythonTypeInference, AnnotateTypes,
+                                NopythonRewrites, IRLegalization)
+
+from numba.core.compiler_machinery import FunctionPass, PassManager, register_pass
+from numba.experimental import jitclass
+import unittest
+
+
+skip_unsupported = skip_parfors_unsupported
+
+
+# test class for #3700
+@jitclass([('L', types.int32), ('T', types.int32)])
+class ExampleClass3700(object):
+    def __init__(self, n):
+        self.L = n
+        self.T = n + 1
+
+
+# test value for test_global_tuple
+GVAL = (1.2,)
+GVAL2 = (3, 4)
 
-# for parallel tests, marking that Windows with Python 2.7 is not supported
-_windows_py27 = (sys.platform.startswith('win32') and
-                 sys.version_info[:2] == (2, 7))
-_32bit = sys.maxsize <= 2 ** 32
-_reason = 'parfors not supported'
-skip_unsupported = unittest.skipIf(_32bit or _windows_py27, _reason)
 
 class TestEquivSet(TestCase):
 
     """
     Test array_analysis.EquivSet.
     """
-    @tag('important')
     def test_insert_equiv(self):
         s1 = EquivSet()
         s1.insert_equiv('a', 'b')
@@ -40,7 +67,6 @@ class TestEquivSet(TestCase):
         self.assertTrue(s1.is_equiv('a', 'b', 'c', 'd'))
         self.assertFalse(s1.is_equiv('a', 'e'))
 
-    @tag('important')
     def test_intersect(self):
         s1 = EquivSet()
         s2 = EquivSet()
@@ -63,7 +89,26 @@ class TestEquivSet(TestCase):
         self.assertFalse(r.is_equiv('c', 'd'))
 
 
-class ArrayAnalysisTester(Pipeline):
+@register_pass(analysis_only=False, mutates_CFG=True)
+class ArrayAnalysisPass(FunctionPass):
+    _name = "array_analysis_pass"
+
+    def __init__(self):
+        FunctionPass.__init__(self)
+
+    def run_pass(self, state):
+        state.array_analysis = ArrayAnalysis(state.typingctx, state.func_ir,
+                                             state.typemap, state.calltypes)
+        state.array_analysis.run(state.func_ir.blocks)
+        post_proc = postproc.PostProcessor(state.func_ir)
+        post_proc.run()
+        state.func_ir_copies.append(state.func_ir.copy())
+        if state.test_idempotence and len(state.func_ir_copies) > 1:
+            state.test_idempotence(state.func_ir_copies)
+        return False
+
+
+class ArrayAnalysisTester(Compiler):
 
     @classmethod
     def mk_pipeline(cls, args, return_type=None, flags=None, locals={},
@@ -82,53 +127,46 @@ class ArrayAnalysisTester(Pipeline):
         """
         Populate and run compiler pipeline
         """
-        self.func_id = bytecode.FunctionIdentity.from_function(func)
+        self.state.func_id = bytecode.FunctionIdentity.from_function(func)
+        ExtractByteCode().run_pass(self.state)
 
-        try:
-            bc = self.extract_bytecode(self.func_id)
-        except BaseException as e:
-            raise e
+        self.state.lifted = ()
+        self.state.lifted_from = None
+        state = self.state
+        state.func_ir_copies = []
+        state.test_idempotence = test_idempotence
 
-        self.bc = bc
-        self.lifted = ()
-        self.lifted_from = None
+        name = 'array_analysis_testing'
+        pm = PassManager(name)
+        pm.add_pass(TranslateByteCode, "analyzing bytecode")
+        pm.add_pass(FixupArgs, "fix up args")
+        pm.add_pass(IRProcessing, "processing IR")
+        # pre typing
+        if not state.flags.no_rewrites:
+            pm.add_pass(GenericRewrites, "nopython rewrites")
+            pm.add_pass(RewriteSemanticConstants, "rewrite semantic constants")
+            pm.add_pass(DeadBranchPrune, "dead branch pruning")
+        pm.add_pass(InlineClosureLikes,
+                    "inline calls to locally defined closures")
+        # typing
+        pm.add_pass(NopythonTypeInference, "nopython frontend")
 
-        pm = _PipelineManager()
+        if not state.flags.no_rewrites:
+            pm.add_pass(NopythonRewrites, "nopython rewrites")
 
-        pm.create_pipeline("nopython")
-        if self.func_ir is None:
-            pm.add_stage(self.stage_analyze_bytecode, "analyzing bytecode")
-        pm.add_stage(self.stage_process_ir, "processing IR")
-        if not self.flags.no_rewrites:
-            if self.status.can_fallback:
-                pm.add_stage(
-                    self.stage_preserve_ir, "preserve IR for fallback")
-            pm.add_stage(self.stage_generic_rewrites, "nopython rewrites")
-        pm.add_stage(
-            self.stage_inline_pass, "inline calls to locally defined closures")
-        pm.add_stage(self.stage_nopython_frontend, "nopython frontend")
-        pm.add_stage(self.stage_annotate_type, "annotate type")
-        if not self.flags.no_rewrites:
-            pm.add_stage(self.stage_nopython_rewrites, "nopython rewrites")
-        func_ir_copies = []
-
-        def stage_array_analysis():
-            self.array_analysis = ArrayAnalysis(self.typingctx, self.func_ir,
-                                                self.type_annotation.typemap,
-                                                self.type_annotation.calltypes)
-            self.array_analysis.run(self.func_ir.blocks)
-            func_ir_copies.append(self.func_ir.copy())
-            if test_idempotence and len(func_ir_copies) > 1:
-                test_idempotence(func_ir_copies)
-
-        pm.add_stage(stage_array_analysis, "analyze array equivalences")
+        # Array Analysis pass
+        pm.add_pass(ArrayAnalysisPass, "array analysis")
         if test_idempotence:
-            # Do another pass of array analysis to test idempontence
-            pm.add_stage(stage_array_analysis, "analyze array equivalences")
+            # Do another pass of array analysis to test idempotence
+            pm.add_pass(ArrayAnalysisPass, "idempotence array analysis")
+        # legalise
+        pm.add_pass(IRLegalization, "ensure IR is legal prior to lowering")
+        pm.add_pass(AnnotateTypes, "annotate types")
 
+        # partial compile
         pm.finalize()
-        res = pm.run(self.status)
-        return self.array_analysis
+        pm.run(state)
+        return state.array_analysis
 
 
 class TestArrayAnalysis(TestCase):
@@ -137,7 +175,7 @@ class TestArrayAnalysis(TestCase):
         outputs = []
         for func_ir in ir_list:
             remove_dead(func_ir.blocks, func_ir.arg_names, func_ir)
-            output = utils.StringIO()
+            output = StringIO()
             func_ir.dump(file=output)
             outputs.append(output.getvalue())
         self.assertTrue(len(set(outputs)) == 1)  # assert all outputs are equal
@@ -153,7 +191,7 @@ class TestArrayAnalysis(TestCase):
             for func in equivs:
                 # only test the equiv_set of the first block
                 func(analysis.equiv_sets[0])
-        if asserts == None:
+        if asserts is None:
             self.assertTrue(self._has_no_assertcall(analysis.func_ir))
         else:
             for func in asserts:
@@ -166,7 +204,7 @@ class TestArrayAnalysis(TestCase):
                 fn = func_ir.get_definition(expr.func.name)
                 if isinstance(fn, ir.Global) and fn.name == 'assert_equiv':
                     typ = typemap[expr.args[0].name]
-                    if typ.value.startswith(msg):
+                    if typ.literal_value.startswith(msg):
                         return True
         return False
 
@@ -327,6 +365,29 @@ class TestArrayAnalysis(TestCase):
                                asserts=[self.with_assert('A', 'B'),
                                         self.without_assert('C', 'D')])
 
+        def test_11():
+            a = np.ones(5)
+            b = np.ones(5)
+            c = a[1:]
+            d = b[:-1]
+            e = len(c)
+            f = len(d)
+            return e == f
+        self._compile_and_test(test_11, (),
+                               equivs=[self.with_equiv('e', 'f')])
+
+        def test_12():
+            a = np.ones(25).reshape((5,5))
+            b = np.ones(25).reshape((5,5))
+            c = a[1:,:]
+            d = b[:-1,:]
+            e = c.shape[0]
+            f = d.shape[0]
+            g = len(d)
+            return e == f
+        self._compile_and_test(test_12, (),
+                               equivs=[self.with_equiv('e', 'f', 'g')])
+
         def test_tup_arg(T):
             T2 = T
             return T2[0]
@@ -335,6 +396,15 @@ class TestArrayAnalysis(TestCase):
         self._compile_and_test(test_tup_arg,
             (types.Tuple((int_arr_typ, int_arr_typ)),), asserts=None)
 
+        def test_arr_in_tup(m):
+            A = np.ones(m)
+            S = (A,)
+            B = np.ones(len(S[0]))
+            return B
+
+        self._compile_and_test(test_arr_in_tup, (types.intp,),
+                               equivs=[self.with_equiv("A", "B")])
+
         T = namedtuple("T", ['a','b'])
         def test_namedtuple(n):
             r = T(n, n)
@@ -405,7 +475,7 @@ class TestArrayAnalysis(TestCase):
 
 
     def test_stencilcall(self):
-        from numba import stencil
+        from numba.stencils.stencil import stencil
         @stencil
         def kernel_1(a):
             return 0.25 * (a[0,1] + a[1,0] + a[0,-1] + a[-1,0])
@@ -524,7 +594,7 @@ class TestArrayAnalysis(TestCase):
             E = B + C
             return E
         self._compile_and_test(test_7, (types.intp,),
-                               asserts=[self.without_assert('B', 'C')],
+                               asserts=[self.with_assert('B', 'C')],
                                idempotent=False)
 
         def test_8(m):
@@ -537,6 +607,25 @@ class TestArrayAnalysis(TestCase):
                                asserts=[self.without_assert('B', 'C')],
                                idempotent=False)
 
+        def test_9(m):
+            # issues #3461 and #3554, checks equivalence on empty slices
+            # and across binop
+            A = np.zeros((m))
+            B = A[:0] # B = array([], dtype=int64)
+            C = A[1:]
+            D = A[:-1:-1] # D = array([], dtype=int64)
+            E = B + D
+            F = E
+            F += 1 # F = array([], dtype=int64)
+            return A, C, F
+        self._compile_and_test(test_9, (types.intp,),
+                               equivs=[self.without_equiv('B', 'C'),
+                                       self.with_equiv('A', 'm'),
+                                       self.with_equiv('B', 'D'),
+                                       self.with_equiv('F', 'D'),],
+                               idempotent=False)
+
+    @skip_unless_scipy
     def test_numpy_calls(self):
         def test_zeros(n):
             a = np.zeros(n)
@@ -782,29 +871,28 @@ class TestArrayAnalysis(TestCase):
                                        self.with_equiv('v', (2, 3, 8)),
                                        ])
 
-        if numpy_version >= (1, 10):
-            def test_stack(m, n):
-                a = np.ones(m)
-                b = np.ones(n)
-                c = np.stack((a, b))
-                d = np.ones((m, n))
-                e = np.ones((m, n))
-                f = np.stack((d, e))
-                g = np.stack((d, e), axis=0)
-                h = np.stack((d, e), axis=1)
-                i = np.stack((d, e), axis=2)
-                j = np.stack((d, e), axis=-1)
+        def test_stack(m, n):
+            a = np.ones(m)
+            b = np.ones(n)
+            c = np.stack((a, b))
+            d = np.ones((m, n))
+            e = np.ones((m, n))
+            f = np.stack((d, e))
+            g = np.stack((d, e), axis=0)
+            h = np.stack((d, e), axis=1)
+            i = np.stack((d, e), axis=2)
+            j = np.stack((d, e), axis=-1)
 
-            self._compile_and_test(test_stack, (types.intp, types.intp),
-                                   equivs=[self.with_equiv('m', 'n'),
-                                           self.with_equiv('c', (2, 'm')),
-                                           self.with_equiv(
-                                       'f', 'g', (2, 'm', 'n')),
-                self.with_equiv(
-                                       'h', ('m', 2, 'n')),
-                self.with_equiv(
-                                       'i', 'j', ('m', 'n', 2)),
-            ])
+        self._compile_and_test(test_stack, (types.intp, types.intp),
+                                equivs=[self.with_equiv('m', 'n'),
+                                        self.with_equiv('c', (2, 'm')),
+                                        self.with_equiv(
+                                    'f', 'g', (2, 'm', 'n')),
+            self.with_equiv(
+                                    'h', ('m', 2, 'n')),
+            self.with_equiv(
+                                    'i', 'j', ('m', 'n', 2)),
+        ])
 
         def test_linspace(m, n):
             a = np.linspace(m, n)
@@ -848,6 +936,16 @@ class TestArrayAnalysis(TestCase):
                                equivs=[self.with_equiv('a', 'c', 'e')],
                                asserts=None)
 
+        # make sure shape of a global tuple of ints is handled properly
+        def test_global_tuple():
+            a = np.ones(GVAL2)
+            b = np.ones(GVAL2)
+
+        self._compile_and_test(test_global_tuple, (),
+                               equivs=[self.with_equiv('a', 'b')],
+                               asserts=None)
+
+
 class TestArrayAnalysisParallelRequired(TestCase):
     """This is to just split out tests that need the parallel backend and
     therefore serialised execution.
@@ -873,5 +971,159 @@ class TestArrayAnalysisParallelRequired(TestCase):
         except IndexError:
             self.fail("test_bug2537 raised IndexError!")
 
+    @skip_unsupported
+    def test_global_namedtuple(self):
+        Row = namedtuple('Row', ['A'])
+        row = Row(3)
+
+        def test_impl():
+            rr = row
+            res = rr.A
+            if res == 2:
+                res = 3
+            return res
+
+        self.assertEqual(njit(test_impl, parallel=True)(), test_impl())
+
+    @skip_unsupported
+    def test_array_T_issue_3700(self):
+
+        def test_impl(t_obj, X):
+            for i in prange(t_obj.T):
+                X[i] = i
+            return X.sum()
+
+        n = 5
+        t_obj = ExampleClass3700(n)
+        X1 = np.zeros(t_obj.T)
+        X2 = np.zeros(t_obj.T)
+        self.assertEqual(
+            njit(test_impl, parallel=True)(t_obj, X1), test_impl(t_obj, X2))
+
+    @skip_unsupported
+    def test_slice_shape_issue_3380(self):
+        # these tests shouldn't throw error in array analysis
+        def test_impl1():
+            a = slice(None, None)
+            return True
+
+        self.assertEqual(njit(test_impl1, parallel=True)(), test_impl1())
+
+        def test_impl2(A, a):
+            b = a
+            return A[b]
+
+        A = np.arange(10)
+        a = slice(None)
+        np.testing.assert_array_equal(
+            njit(test_impl2, parallel=True)(A, a), test_impl2(A, a))
+
+    @skip_unsupported
+    def test_slice_dtype_issue_5056(self):
+        # see issue 5056
+
+        @njit(parallel=True)
+        def test_impl(data):
+            N = data.shape[0]
+            sums = np.zeros(N)
+            for i in prange(N):
+                sums[i] = np.sum(data[np.int32(0):np.int32(1)])
+            return sums
+
+        data = np.arange(10.)
+        np.testing.assert_array_equal(test_impl(data), test_impl.py_func(data))
+
+    @skip_unsupported
+    def test_global_tuple(self):
+        """make sure a global tuple with non-integer values does not cause errors
+        (test for #6726).
+        """
+
+        def test_impl():
+            d = GVAL[0]
+            return d
+
+        self.assertEqual(njit(test_impl, parallel=True)(), test_impl())
+
+
+class TestArrayAnalysisInterface(TestCase):
+    def test_analyze_op_call_interface(self):
+        # gather _analyze_op_call_*
+        aoc = {}
+        for fname in dir(ArrayAnalysis):
+            if fname.startswith('_analyze_op_call_'):
+                aoc[fname] = getattr(ArrayAnalysis, fname)
+        # check interface
+        def iface_stub(self, scope, equiv_set, loc, args, kws):
+            pass
+        expected = utils.pysignature(iface_stub)
+        for k, v in aoc.items():
+            got = utils.pysignature(v)
+            with self.subTest(fname=k, sig=got):
+                self.assertEqual(got, expected)
+
+    @skip_unsupported
+    def test_array_analysis_extensions(self):
+        # Test that the `array_analysis` object in `array_analysis_extensions`
+        # can perform analysis on the scope using `equiv_sets`.
+        from numba.parfors.parfor import Parfor
+        from numba.parfors import array_analysis
+
+        orig_parfor = array_analysis.array_analysis_extensions[Parfor]
+
+        shared = {'counter': 0}
+
+        def testcode(array_analysis):
+            # Find call node corresponding to the ``A = empty(n)``
+            func_ir = array_analysis.func_ir
+            for call in func_ir.blocks[0].find_exprs('call'):
+                callee = func_ir.get_definition(call.func)
+                if getattr(callee, "value", None) is empty:
+                    if getattr(call.args[0], 'name', None) == 'n':
+                        break
+            else:
+                return
+
+            variable_A = func_ir.get_assignee(call)
+            # n must be equiv to
+            es = array_analysis.equiv_sets[0]
+            self.assertTrue(es.is_equiv('n', variable_A.name))
+            shared['counter'] += 1
+
+        def new_parfor(parfor, equiv_set, typemap, array_analysis):
+            """Recursive array analysis for parfor nodes.
+            """
+            testcode(array_analysis)
+            # Call original
+            return orig_parfor(
+                parfor, equiv_set, typemap, array_analysis,
+            )
+
+        try:
+            # Replace the array-analysis extension for Parfor node
+            array_analysis.array_analysis_extensions[Parfor] = new_parfor
+
+            empty = np.empty   # avoid scanning a getattr in the IR
+            def f(n):
+                A = empty(n)
+                for i in prange(n):
+                    S = np.arange(i)
+                    A[i] = S.sum()
+                return A + 1
+
+            got = njit(parallel=True)(f)(10)
+            executed_count = shared['counter']
+            self.assertGreater(executed_count, 0)
+        finally:
+            # Re-install the original handler
+            array_analysis.array_analysis_extensions[Parfor] = orig_parfor
+
+        # Check normal execution
+        expected = njit(parallel=True)(f)(10)
+        self.assertPreciseEqual(got, expected)
+        # Make sure we have uninstalled the handler
+        self.assertEqual(executed_count, shared['counter'])
+
+
 if __name__ == '__main__':
     unittest.main()
