@@ -1,20 +1,23 @@
-from __future__ import print_function
-
 import collections
 import sys
 import weakref
+import gc
 
-import numba.unittest_support as unittest
-from numba.controlflow import CFGraph, Loop
-from numba.compiler import compile_extra, compile_isolated, Flags
-from numba import types
-from .support import TestCase
+import unittest
+from numba import njit
+from numba.core.controlflow import CFGraph, Loop
+from numba.core.compiler import (compile_extra, compile_isolated, Flags,
+                                 CompilerBase, DefaultPassBuilder)
+from numba.core.untyped_passes import PreserveIR
+from numba.core.typed_passes import IRLegalization
+from numba.core import types, ir
+from numba.tests.support import TestCase, override_config, SerialMixin
 
 enable_pyobj_flags = Flags()
-enable_pyobj_flags.set("enable_pyobject")
+enable_pyobj_flags.enable_pyobject = True
 
 forceobj_flags = Flags()
-forceobj_flags.set("force_pyobject")
+forceobj_flags.force_pyobject = True
 
 no_pyobj_flags = Flags()
 
@@ -223,6 +226,20 @@ def del_before_definition(rec):
     return -1
 
 
+def inf_loop_multiple_back_edge(rec):
+    """
+    test to reveal bug of invalid liveness when infinite loop has multiple
+    backedge.
+    """
+    while True:
+        rec.mark("yield")
+        yield
+        p = rec('p')
+        if p:
+            rec.mark('bra')
+            pass
+
+
 class TestObjLifetime(TestCase):
     """
     Test lifetime of Python objects inside jit-compiled functions.
@@ -320,6 +337,7 @@ class TestObjLifetime(TestCase):
         with self.assertRefCount(rec):
             gen = cfunc(rec)
             del gen
+            gc.collect()
             self.assertFalse(rec.alive)
         # Stop iterating before exhaustion
         rec = RefRecorder()
@@ -328,6 +346,7 @@ class TestObjLifetime(TestCase):
             next(gen)
             self.assertTrue(rec.alive)
             del gen
+            gc.collect()
             self.assertFalse(rec.alive)
 
     def test_generator1(self):
@@ -355,6 +374,75 @@ class TestObjLifetime(TestCase):
             rec = self.compile_and_record(raising_usecase3, raises=MyError)
             self.assertFalse(rec.alive)
 
+    def test_inf_loop_multiple_back_edge(self):
+        cfunc = self.compile(inf_loop_multiple_back_edge)
+        rec = RefRecorder()
+        iterator = iter(cfunc(rec))
+        next(iterator)
+        self.assertEqual(rec.alive, [])
+        next(iterator)
+        self.assertEqual(rec.alive, [])
+        next(iterator)
+        self.assertEqual(rec.alive, [])
+        self.assertEqual(rec.recorded,
+                         ['yield', 'p', 'bra', 'yield', 'p', 'bra', 'yield'])
+
+
+class TestExtendingVariableLifetimes(SerialMixin, TestCase):
+    # Test for `numba.config.EXTEND_VARIABLE_LIFETIMES` which moves the ir.Del
+    # nodes to just before a block's terminator, i.e. their lifetime is extended
+    # beyond the point of last use.
+
+    def test_lifetime_basic(self):
+
+        def get_ir(extend_lifetimes):
+            class IRPreservingCompiler(CompilerBase):
+
+                def define_pipelines(self):
+                    pm = DefaultPassBuilder.define_nopython_pipeline(self.state)
+                    pm.add_pass_after(PreserveIR, IRLegalization)
+                    pm.finalize()
+                    return [pm]
+
+            @njit(pipeline_class=IRPreservingCompiler)
+            def foo():
+                a = 10
+                b = 20
+                c = a + b
+                # a and b are now unused, standard behaviour is ir.Del for them here
+                d = c / c
+                return d
+
+            with override_config('EXTEND_VARIABLE_LIFETIMES', extend_lifetimes):
+                foo()
+                cres = foo.overloads[foo.signatures[0]]
+                func_ir = cres.metadata['preserved_ir']
+
+            return func_ir
+
+
+        def check(func_ir, expect):
+            # assert single block
+            self.assertEqual(len(func_ir.blocks), 1)
+            blk = next(iter(func_ir.blocks.values()))
+
+            # check sequencing
+            for expect_class, got_stmt in zip(expect, blk.body):
+                self.assertIsInstance(got_stmt, expect_class)
+
+        del_after_use_ir = get_ir(False)
+        # should be 3 assigns (a, b, c), 2 del (a, b), assign (d), del (c)
+        # assign for cast d to return, del (d), return
+        expect = [*((ir.Assign,) * 3), ir.Del, ir.Del, ir.Assign, ir.Del,
+                  ir.Assign, ir.Del, ir.Return]
+        check(del_after_use_ir, expect)
+
+        del_at_block_end_ir = get_ir(True)
+        # should be 4 assigns (a, b, c, d), assign for cast d to return,
+        # 4 dels (a, b, c, d) then the return.
+        expect = [*((ir.Assign,) * 4), ir.Assign, *((ir.Del,) * 4), ir.Return]
+        check(del_at_block_end_ir, expect)
+
 
 if __name__ == "__main__":
     unittest.main()
