@@ -1,6 +1,4 @@
-from __future__ import print_function
-
-import numba.unittest_support as unittest
+import unittest
 
 from collections import namedtuple
 import contextlib
@@ -11,11 +9,12 @@ import sys
 
 import numpy as np
 
-from numba.compiler import compile_isolated, Flags
-from numba import jit, types
-import numba.unittest_support as unittest
-from .support import (TestCase, enable_pyobj_flags, MemoryLeakMixin, tag,
-                      compile_function)
+from numba.core.compiler import compile_isolated, Flags, errors
+from numba import jit
+from numba.core import types
+import unittest
+from numba.tests.support import (TestCase, enable_pyobj_flags, MemoryLeakMixin,
+                                 tag, compile_function)
 
 
 Point = namedtuple('Point', ('a', 'b'))
@@ -64,6 +63,11 @@ def update_usecase(a, b, c):
     s.update(c)
     return list(s)
 
+def bool_usecase(arg):
+    # Remove one element to allow for empty sets.
+    s = set(arg[1:])
+    return bool(s)
+
 def remove_usecase(a, b):
     s = set(a)
     for v in b:
@@ -275,10 +279,6 @@ def unique_usecase(src):
     return res
 
 
-needs_set_literals = unittest.skipIf(sys.version_info < (2, 7),
-                                     "set literals unavailable before Python 2.7")
-
-
 class BaseTest(MemoryLeakMixin, TestCase):
 
     def setUp(self):
@@ -335,23 +335,18 @@ class BaseTest(MemoryLeakMixin, TestCase):
 
 class TestSetLiterals(BaseTest):
 
-    @needs_set_literals
     def test_build_set(self, flags=enable_pyobj_flags):
         pyfunc = set_literal_return_usecase((1, 2, 3, 2))
         self.run_nullary_func(pyfunc, flags=flags)
 
-    @needs_set_literals
-    def test_build_heterogenous_set(self, flags=enable_pyobj_flags):
+    def test_build_heterogeneous_set(self, flags=enable_pyobj_flags):
         pyfunc = set_literal_return_usecase((1, 2.0, 3j, 2))
         self.run_nullary_func(pyfunc, flags=flags)
-        # Check that items are inserted in the right order (here the
-        # result will be {2}, not {2.0})
         pyfunc = set_literal_return_usecase((2.0, 2))
         got, expected = self.run_nullary_func(pyfunc, flags=flags)
+
         self.assertIs(type(got.pop()), type(expected.pop()))
 
-    @tag('important')
-    @needs_set_literals
     def test_build_set_nopython(self):
         arg = list(self.sparse_array(50))
         pyfunc = set_literal_convert_usecase(arg)
@@ -378,7 +373,6 @@ class TestSets(BaseTest):
         check(self.duplicates_array(200))
         check(self.sparse_array(200))
 
-    @tag('important')
     def test_set_return(self):
         pyfunc = set_return_usecase
         cfunc = jit(nopython=True)(pyfunc)
@@ -386,7 +380,6 @@ class TestSets(BaseTest):
         arg = (1, 2, 3, 2, 7)
         self.assertEqual(cfunc(arg), set(arg))
 
-    @tag('important')
     def test_iterator(self):
         pyfunc = iterator_usecase
         check = self.unordered_checker(pyfunc)
@@ -395,7 +388,6 @@ class TestSets(BaseTest):
         check(self.duplicates_array(200))
         check(self.sparse_array(200))
 
-    @tag('important')
     def test_update(self):
         pyfunc = update_usecase
         check = self.unordered_checker(pyfunc)
@@ -408,6 +400,15 @@ class TestSets(BaseTest):
         c = self.sparse_array(50)
         check(a, b, c)
 
+    def test_bool(self):
+        pyfunc = bool_usecase
+        check = self.unordered_checker(pyfunc)
+
+        check([1])
+        check([1, 2])
+        check([False, False])
+        check([True, False])
+
     def test_remove(self):
         pyfunc = remove_usecase
         check = self.unordered_checker(pyfunc)
@@ -425,7 +426,19 @@ class TestSets(BaseTest):
         with self.assertRaises(KeyError) as raises:
             cfunc((1, 2, 3), (5, ))
 
-    @tag('important')
+    def test_refcounted_types_forbidden(self):
+        # References are leaked on exception
+        self.disable_leak_check()
+
+        pyfunc = constructor_usecase
+        cfunc = jit(nopython=True)(pyfunc)
+        with self.assertRaises(errors.NumbaValueError) as raises:
+            cfunc("abc")
+
+        excstr = str(raises.exception)
+        self.assertIn("Use of reference counted items in 'set()'", excstr)
+        self.assertIn("offending type is: 'unicode_type'", excstr)
+
     def test_discard(self):
         pyfunc = discard_usecase
         check = self.unordered_checker(pyfunc)
@@ -448,7 +461,6 @@ class TestSets(BaseTest):
         check = self.unordered_checker(pyfunc)
         check((1,), 5, 5)
 
-    @tag('important')
     def test_pop(self):
         pyfunc = pop_usecase
         check = self.unordered_checker(pyfunc)
@@ -456,7 +468,6 @@ class TestSets(BaseTest):
         check((2, 3, 55, 11, 8, 42))
         check(self.sparse_array(50))
 
-    @tag('important')
     def test_contains(self):
         pyfunc = contains_usecase
         cfunc = jit(nopython=True)(pyfunc)
@@ -622,7 +633,6 @@ class OtherTypesTest(object):
         check(self.duplicates_array(200))
         check(self.sparse_array(200))
 
-    @tag('important')
     def test_update(self):
         pyfunc = update_usecase
         check = self.unordered_checker(pyfunc)
@@ -676,7 +686,6 @@ class TestUnboxing(BaseTest):
             self.assertPreciseEqual(got, expected)
         return check
 
-    @tag('important')
     def test_numbers(self):
         check = self.check_unary(unbox_usecase)
         check(set([1, 2]))
@@ -689,7 +698,6 @@ class TestUnboxing(BaseTest):
         check(set([(1, 2), (3, 4)]))
         check(set([(1, 2j), (3, 4j)]))
 
-    @tag('important')
     def test_set_inside_tuple(self):
         check = self.check_unary(unbox_usecase3)
         check((1, set([2, 3, 4])))
@@ -699,8 +707,8 @@ class TestUnboxing(BaseTest):
         check((1, set([(2,), (3,)])))
 
     def test_errors(self):
-        # Error checking should ensure the set is homogenous
-        msg = "can't unbox heterogenous set"
+        # Error checking should ensure the set is homogeneous
+        msg = "can't unbox heterogeneous set"
         pyfunc = noop
         cfunc = jit(nopython=True)(pyfunc)
         val = set([1, 2.5])
@@ -764,7 +772,6 @@ class TestSetReflection(BaseTest):
                 cfunc(s)
             self.assertPreciseEqual(s, set([1, 2, 3, 42]))
 
-    @tag('important')
     def test_reflect_same_set(self):
         """
         When the same set object is reflected twice, behaviour should
@@ -797,7 +804,6 @@ class TestExamples(BaseTest):
     Examples of using sets.
     """
 
-    @tag('important')
     def test_unique(self):
         pyfunc = unique_usecase
         check = self.unordered_checker(pyfunc)
@@ -805,6 +811,17 @@ class TestExamples(BaseTest):
         check(self.duplicates_array(200))
         check(self.sparse_array(200))
 
+    def test_type_coercion_from_update(self):
+        # see issue #6621
+        def impl():
+            i = np.uint64(1)
+            R = set()
+            R.update({1, 2, 3})
+            R.add(i)
+            return R
+        check = self.unordered_checker(impl)
+        check()
+
 
 if __name__ == '__main__':
     unittest.main()
