@@ -1,5 +1,3 @@
-from __future__ import print_function
-
 import copy
 import itertools
 import math
@@ -8,14 +6,16 @@ import sys
 
 import numpy as np
 
-from numba.compiler import compile_isolated, Flags
-from numba import jit, types, utils
-import numba.unittest_support as unittest
+from numba.core.compiler import compile_isolated, Flags
+from numba import jit, njit
+from numba.core import types, utils, errors
+import unittest
 from numba import testing
-from .support import TestCase, MemoryLeakMixin, tag
+from numba.tests.support import TestCase, MemoryLeakMixin, tag
 
-from numba.targets.quicksort import make_py_quicksort, make_jit_quicksort
-from .timsort import make_py_timsort, make_jit_timsort, MergeRun
+from numba.misc.quicksort import make_py_quicksort, make_jit_quicksort
+from numba.misc.mergesort import make_jit_mergesort
+from numba.misc.timsort import make_py_timsort, make_jit_timsort, MergeRun
 
 
 def make_temp_list(keys, n):
@@ -41,6 +41,15 @@ jit_quicksort = make_jit_quicksort()
 def sort_usecase(val):
     val.sort()
 
+def argsort_usecase(val):
+    return val.argsort()
+
+def argsort_kind_usecase(val, is_stable=False):
+    if is_stable:
+        return val.argsort(kind='mergesort')
+    else:
+        return val.argsort(kind='quicksort')
+
 def sorted_usecase(val):
     return sorted(val)
 
@@ -50,6 +59,15 @@ def sorted_reverse_usecase(val, b):
 def np_sort_usecase(val):
     return np.sort(val)
 
+def np_argsort_usecase(val):
+    return np.argsort(val)
+
+def np_argsort_kind_usecase(val, is_stable=False):
+    if is_stable:
+        return np.argsort(val, kind='mergesort')
+    else:
+        return np.argsort(val, kind='quicksort')
+
 def list_sort_usecase(n):
     np.random.seed(42)
     l = []
@@ -155,7 +173,6 @@ class BaseTimsortTest(BaseSortingTest):
         f = self.timsort.merge_init
         return f(keys)
 
-    @tag('important')
     def test_binarysort(self):
         n = 20
         def check(l, n, start=0):
@@ -241,7 +258,6 @@ class BaseTimsortTest(BaseSortingTest):
         for i in range(len(l) - 1):
             check(l, i, n)
 
-    @tag('important')
     def test_gallop_left(self):
         n = 20
         f = self.timsort.gallop_left
@@ -548,11 +564,13 @@ class TestTimsortArrays(JITTimsortMixin, BaseTimsortTest, TestCase):
 
 class BaseQuicksortTest(BaseSortingTest):
 
+    # NOTE these tests assume a non-argsort quicksort.
+
     def test_insertion_sort(self):
         n = 20
         def check(l, n):
             res = self.array_factory([9999] + l + [-9999])
-            f(res, 1, n)
+            f(res, res, 1, n)
             self.assertEqual(res[0], 9999)
             self.assertEqual(res[-1], -9999)
             self.assertSorted(l, res[1:-1])
@@ -575,7 +593,7 @@ class BaseQuicksortTest(BaseSortingTest):
         n = 20
         def check(l, n):
             res = self.array_factory([9999] + l + [-9999])
-            index = f(res, 1, n)
+            index = f(res, res, 1, n)
             self.assertEqual(res[0], 9999)
             self.assertEqual(res[-1], -9999)
             pivot = res[index]
@@ -628,7 +646,6 @@ class BaseQuicksortTest(BaseSortingTest):
         l = self.duprandom_list(n)
         check(l, n)
 
-    @tag('important')
     def test_run_quicksort(self):
         f = self.quicksort.run_quicksort
 
@@ -704,6 +721,29 @@ class TestNumpySort(TestCase):
     def setUp(self):
         np.random.seed(42)
 
+    def int_arrays(self):
+        for size in (5, 20, 50, 500):
+            yield np.random.randint(99, size=size)
+
+    def float_arrays(self):
+        for size in (5, 20, 50, 500):
+            yield np.random.random(size=size) * 100
+        # Now with NaNs.  Numpy sorts them at the end.
+        for size in (5, 20, 50, 500):
+            orig = np.random.random(size=size) * 100
+            orig[np.random.random(size=size) < 0.1] = float('nan')
+            yield orig
+
+    def has_duplicates(self, arr):
+        """
+        Whether the array has duplicates.  Takes NaNs into account.
+        """
+        if np.count_nonzero(np.isnan(arr)) > 1:
+            return True
+        if np.unique(arr).size < arr.size:
+            return True
+        return False
+
     def check_sort_inplace(self, pyfunc, cfunc, val):
         expected = copy.copy(val)
         got = copy.copy(val)
@@ -719,35 +759,38 @@ class TestNumpySort(TestCase):
         # The original wasn't mutated
         self.assertPreciseEqual(val, orig)
 
+    def check_argsort(self, pyfunc, cfunc, val, kwargs={}):
+        orig = copy.copy(val)
+        expected = pyfunc(val, **kwargs)
+        got = cfunc(val, **kwargs)
+        self.assertPreciseEqual(orig[got], np.sort(orig),
+                                msg="the array wasn't argsorted")
+        # Numba and Numpy results may differ if there are duplicates
+        # in the array
+        if not self.has_duplicates(orig):
+            self.assertPreciseEqual(got, expected)
+        # The original wasn't mutated
+        self.assertPreciseEqual(val, orig)
+
     def test_array_sort_int(self):
         pyfunc = sort_usecase
         cfunc = jit(nopython=True)(pyfunc)
 
-        for size in (5, 20, 50, 500):
-            orig = np.random.randint(99, size=size)
+        for orig in self.int_arrays():
             self.check_sort_inplace(pyfunc, cfunc, orig)
 
-    @tag('important')
     def test_array_sort_float(self):
         pyfunc = sort_usecase
         cfunc = jit(nopython=True)(pyfunc)
 
-        for size in (5, 20, 50, 500):
-            orig = np.random.random(size=size) * 100
-            self.check_sort_inplace(pyfunc, cfunc, orig)
-
-        # Now with NaNs.  Numpy sorts them at the end.
-        for size in (5, 20, 50, 500):
-            orig = np.random.random(size=size) * 100
-            orig[np.random.random(size=size) < 0.1] = float('nan')
+        for orig in self.float_arrays():
             self.check_sort_inplace(pyfunc, cfunc, orig)
 
     def test_np_sort_int(self):
         pyfunc = np_sort_usecase
         cfunc = jit(nopython=True)(pyfunc)
 
-        for size in (5, 20, 50, 500):
-            orig = np.random.randint(99, size=size)
+        for orig in self.int_arrays():
             self.check_sort_copy(pyfunc, cfunc, orig)
 
     def test_np_sort_float(self):
@@ -759,10 +802,51 @@ class TestNumpySort(TestCase):
             orig[np.random.random(size=size) < 0.1] = float('nan')
             self.check_sort_copy(pyfunc, cfunc, orig)
 
+    def test_argsort_int(self):
+        def check(pyfunc):
+            cfunc = jit(nopython=True)(pyfunc)
+            for orig in self.int_arrays():
+                self.check_argsort(pyfunc, cfunc, orig)
+
+        check(argsort_usecase)
+        check(np_argsort_usecase)
+
+    def test_argsort_kind_int(self):
+        def check(pyfunc, is_stable):
+            cfunc = jit(nopython=True)(pyfunc)
+            for orig in self.int_arrays():
+                self.check_argsort(pyfunc, cfunc, orig,
+                                   dict(is_stable=is_stable))
+
+        check(argsort_kind_usecase, is_stable=True)
+        check(np_argsort_kind_usecase, is_stable=True)
+        check(argsort_kind_usecase, is_stable=False)
+        check(np_argsort_kind_usecase, is_stable=False)
+
+    def test_argsort_float(self):
+        def check(pyfunc):
+            cfunc = jit(nopython=True)(pyfunc)
+            for orig in self.float_arrays():
+                self.check_argsort(pyfunc, cfunc, orig)
+
+        check(argsort_usecase)
+        check(np_argsort_usecase)
+
+    def test_argsort_float(self):
+        def check(pyfunc, is_stable):
+            cfunc = jit(nopython=True)(pyfunc)
+            for orig in self.float_arrays():
+                self.check_argsort(pyfunc, cfunc, orig,
+                                   dict(is_stable=is_stable))
+
+        check(argsort_kind_usecase, is_stable=True)
+        check(np_argsort_kind_usecase, is_stable=True)
+        check(argsort_kind_usecase, is_stable=False)
+        check(np_argsort_kind_usecase, is_stable=False)
+
 
 class TestPythonSort(TestCase):
 
-    @tag('important')
     def test_list_sort(self):
         pyfunc = list_sort_usecase
         cfunc = jit(nopython=True)(pyfunc)
@@ -806,5 +890,212 @@ class TestPythonSort(TestCase):
             self.assertNotEqual(list(orig), got)   # sanity check
 
 
+class TestMergeSort(TestCase):
+    def setUp(self):
+        np.random.seed(321)
+
+    def check_argsort_stable(self, sorter, low, high, count):
+        # make data with high possibility of duplicated key
+        data = np.random.randint(low, high, count)
+        expect = np.argsort(data, kind='mergesort')
+        got = sorter(data)
+        np.testing.assert_equal(expect, got)
+
+    def test_argsort_stable(self):
+        arglist = [
+            (-2, 2, 5),
+            (-5, 5, 10),
+            (0, 10, 101),
+            (0, 100, 1003),
+        ]
+        imp = make_jit_mergesort(is_argsort=True)
+        toplevel = imp.run_mergesort
+        sorter = njit(lambda arr: toplevel(arr))
+        for args in arglist:
+            self.check_argsort_stable(sorter, *args)
+
+
+nop_compiler = lambda x:x
+
+
+class TestSortSlashSortedWithKey(MemoryLeakMixin, TestCase):
+
+    def test_01(self):
+
+        a = [3, 1, 4, 1, 5, 9]
+
+        @njit
+        def external_key(z):
+            return 1. / z
+
+        @njit
+        def foo(x, key=None):
+            new_x = x[:]
+            new_x.sort(key=key)
+            return sorted(x[:], key=key), new_x
+
+        self.assertPreciseEqual(foo(a[:]), foo.py_func(a[:]))
+        self.assertPreciseEqual(foo(a[:], external_key),
+                                foo.py_func(a[:], external_key))
+
+    def test_02(self):
+
+        a = [3, 1, 4, 1, 5, 9]
+
+        @njit
+        def foo(x):
+            def closure_key(z):
+                return 1. / z
+            new_x = x[:]
+            new_x.sort(key=closure_key)
+            return sorted(x[:], key=closure_key), new_x
+
+        self.assertPreciseEqual(foo(a[:]), foo.py_func(a[:]))
+
+    def test_03(self):
+
+        a = [3, 1, 4, 1, 5, 9]
+
+        def gen(compiler):
+
+            @compiler
+            def bar(x, func):
+                new_x = x[:]
+                new_x.sort(key=func)
+                return sorted(x[:], key=func), new_x
+
+            @compiler
+            def foo(x):
+                def closure_escapee_key(z):
+                    return 1. / z
+                return bar(x, closure_escapee_key)
+
+            return foo
+
+        self.assertPreciseEqual(gen(njit)(a[:]), gen(nop_compiler)(a[:]))
+
+    def test_04(self):
+
+        a = ['a','b','B','b','C','A']
+
+        @njit
+        def external_key(z):
+            return z.upper()
+
+        @njit
+        def foo(x, key=None):
+            new_x = x[:]
+            new_x.sort(key=key)
+            return sorted(x[:], key=key), new_x
+
+        self.assertPreciseEqual(foo(a[:]), foo.py_func(a[:]))
+        self.assertPreciseEqual(foo(a[:], external_key),
+                                foo.py_func(a[:], external_key))
+
+    def test_05(self):
+
+        a = ['a','b','B','b','C','A']
+
+        @njit
+        def external_key(z):
+            return z.upper()
+
+        @njit
+        def foo(x, key=None, reverse=False):
+            new_x = x[:]
+            new_x.sort(key=key, reverse=reverse)
+            return (sorted(x[:], key=key, reverse=reverse), new_x)
+
+        for key, rev in itertools.product((None, external_key),
+                                          (True, False, 1, -12, 0)):
+            self.assertPreciseEqual(foo(a[:], key, rev),
+                                    foo.py_func(a[:], key, rev))
+
+    def test_optional_on_key(self):
+        a = [3, 1, 4, 1, 5, 9]
+
+        @njit
+        def foo(x, predicate):
+            if predicate:
+                def closure_key(z):
+                    return 1. / z
+            else:
+                closure_key = None
+
+            new_x = x[:]
+            new_x.sort(key=closure_key)
+
+            return (sorted(x[:], key=closure_key), new_x)
+
+        with self.assertRaises(errors.TypingError) as raises:
+            TF = True
+            foo(a[:], TF)
+
+        msg = "Key must concretely be None or a Numba JIT compiled function"
+        self.assertIn(msg, str(raises.exception))
+
+    def test_exceptions_sorted(self):
+
+        @njit
+        def foo_sorted(x, key=None, reverse=False):
+            return sorted(x[:], key=key, reverse=reverse)
+
+        @njit
+        def foo_sort(x, key=None, reverse=False):
+            new_x = x[:]
+            new_x.sort(key=key, reverse=reverse)
+            return new_x
+
+        @njit
+        def external_key(z):
+            return 1. / z
+
+        a = [3, 1, 4, 1, 5, 9]
+
+        for impl in (foo_sort, foo_sorted):
+
+            # check illegal key
+            with self.assertRaises(errors.TypingError) as raises:
+                impl(a, key="illegal")
+
+            expect = "Key must be None or a Numba JIT compiled function"
+            self.assertIn(expect, str(raises.exception))
+
+            # check illegal reverse
+            with self.assertRaises(errors.TypingError) as raises:
+                impl(a, key=external_key, reverse="go backwards")
+
+            expect = "an integer is required for 'reverse'"
+            self.assertIn(expect, str(raises.exception))
+
+
+class TestArrayArgsort(MemoryLeakMixin, TestCase):
+    """Tests specific to array.argsort"""
+
+    def test_exceptions(self):
+
+        @njit
+        def nonliteral_kind(kind):
+            np.arange(5).argsort(kind=kind)
+
+        # check non-literal kind
+        with self.assertRaises(errors.TypingError) as raises:
+            # valid spelling but not literal
+            nonliteral_kind('quicksort')
+
+        expect = '"kind" must be a string literal'
+        self.assertIn(expect, str(raises.exception))
+
+        @njit
+        def unsupported_kwarg():
+            np.arange(5).argsort(foo='')
+
+        with self.assertRaises(errors.TypingError) as raises:
+            unsupported_kwarg()
+
+        expect = "Unsupported keywords: ['foo']"
+        self.assertIn(expect, str(raises.exception))
+
+
 if __name__ == '__main__':
     unittest.main()
