@@ -3,7 +3,6 @@
 # SPDX-License-Identifier: BSD-2-Clause
 #
 
-from __future__ import print_function, division, absolute_import
 
 import sys
 import numpy as np
@@ -15,22 +14,17 @@ from contextlib import contextmanager
 from copy import deepcopy
 
 import numba
-from numba import unittest_support as unittest
-from numba import njit, stencil, types
-from numba.compiler import compile_extra, Flags
-from numba.targets import registry
-from numba.targets.cpu import ParallelOptions
-from .support import tag
-from numba.errors import LoweringError, TypingError
+from numba import njit, stencil
+from numba.core.utils import PYVERSION
+from numba.core import types, registry
+from numba.core.compiler import compile_extra, Flags
+from numba.core.cpu import ParallelOptions
+from numba.tests.support import tag, skip_parfors_unsupported, _32bit
+from numba.core.errors import LoweringError, TypingError, NumbaValueError
+import unittest
 
 
-# for decorating tests, marking that Windows with Python 2.7 is not supported
-_py27 = sys.version_info[:2] == (2, 7)
-_windows_py27 = (sys.platform.startswith('win32') and _py27)
-_32bit = sys.maxsize <= 2 ** 32
-_reason = 'parfors not supported'
-_unsupported = _32bit or _windows_py27
-skip_unsupported = unittest.skipIf(_unsupported, _reason)
+skip_unsupported = skip_parfors_unsupported
 
 
 @stencil
@@ -63,6 +57,11 @@ def stencil_multiple_input_kernel_var(a, b, w):
                 b[0, 1] + b[1, 0] + b[0, -1] + b[-1, 0])
 
 
+@stencil
+def stencil_multiple_input_mixed_types_2d(a, b, f):
+    return a[0, 0] if f[0, 0] else b[0, 0]
+
+
 @stencil(standard_indexing=("b",))
 def stencil_with_standard_indexing_1d(a, b):
     return a[-1] * b[0] + a[0] * b[1]
@@ -78,14 +77,14 @@ def stencil_with_standard_indexing_2d(a, b):
 def addone_njit(a):
     return a + 1
 
-# guard against the decorator being run on unsupported platforms
-# as it will raise and stop test discovery from working
-if not _unsupported:
+
+if not _32bit: # prevent compilation on unsupported 32bit targets
     @njit(parallel=True)
     def addone_pjit(a):
         return a + 1
 
 
+@unittest.skipIf(PYVERSION != (3, 7), "Run under 3.7 only, AST unstable")
 class TestStencilBase(unittest.TestCase):
 
     _numba_parallel_test_ = False
@@ -93,7 +92,7 @@ class TestStencilBase(unittest.TestCase):
     def __init__(self, *args):
         # flags for njit()
         self.cflags = Flags()
-        self.cflags.set('nrt')
+        self.cflags.nrt = True
 
         super(TestStencilBase, self).__init__(*args)
 
@@ -104,9 +103,9 @@ class TestStencilBase(unittest.TestCase):
 
     def compile_parallel(self, func, sig, **kws):
         flags = Flags()
-        flags.set('nrt')
+        flags.nrt = True
         options = True if not kws else kws
-        flags.set('auto_parallel', ParallelOptions(options))
+        flags.auto_parallel=ParallelOptions(options)
         return self._compile_this(func, sig, flags)
 
     def compile_njit(self, func, sig):
@@ -147,7 +146,6 @@ class TestStencil(TestStencilBase):
         super(TestStencil, self).__init__(*args, **kwargs)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil1(self):
         """Tests whether the optional out argument to stencil calls works.
         """
@@ -176,7 +174,6 @@ class TestStencil(TestStencilBase):
         self.check(test_impl_seq, test_without_out, n)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil2(self):
         """Tests whether the optional neighborhood argument to the stencil
         decorate works.
@@ -262,7 +259,6 @@ class TestStencil(TestStencilBase):
         self.assertIn('@do_scheduling', cpfunc.library.get_llvm_str())
 
     @skip_unsupported
-    @tag('important')
     def test_stencil3(self):
         """Tests whether a non-zero optional cval argument to the stencil
         decorator works.  Also tests integer result type.
@@ -285,7 +281,6 @@ class TestStencil(TestStencilBase):
         self.assertTrue(par_res[0, 0] == 1.0 and par_res[4, 4] == 1.0)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil_standard_indexing_1d(self):
         """Tests standard indexing with a 1d array.
         """
@@ -308,7 +303,6 @@ class TestStencil(TestStencilBase):
         self.check(test_impl_seq, test_seq, n)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil_standard_indexing_2d(self):
         """Tests standard indexing with a 2d array and multiple stencil calls.
         """
@@ -339,7 +333,6 @@ class TestStencil(TestStencilBase):
         self.check(test_impl_seq, test_seq, n)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil_multiple_inputs(self):
         """Tests whether multiple inputs of the same size work.
         """
@@ -375,7 +368,28 @@ class TestStencil(TestStencilBase):
         self.check(test_impl_seq, test_seq, n)
 
     @skip_unsupported
-    @tag('important')
+    def test_stencil_mixed_types(self):
+        def test_impl_seq(n):
+            A = np.arange(n ** 2).reshape((n, n))
+            B = n ** 2 - np.arange(n ** 2).reshape((n, n))
+            S = np.eye(n, dtype=np.bool_)
+            O = np.zeros((n, n), dtype=A.dtype)
+            for i in range(0, n):
+                for j in range(0, n):
+                    O[i, j] = A[i, j] if S[i, j] else B[i, j]
+            return O
+
+        def test_seq(n):
+            A = np.arange(n ** 2).reshape((n, n))
+            B = n ** 2 - np.arange(n ** 2).reshape((n, n))
+            S = np.eye(n, dtype=np.bool_)
+            O = stencil_multiple_input_mixed_types_2d(A, B, S)
+            return O
+
+        n = 3
+        self.check(test_impl_seq, test_seq, n)
+
+    @skip_unsupported
     def test_stencil_call(self):
         """Tests 2D numba.stencil calls.
         """
@@ -409,7 +423,6 @@ class TestStencil(TestStencilBase):
         self.check(test_impl_seq, test_impl2, n)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil_call_1D(self):
         """Tests 1D numba.stencil calls.
         """
@@ -430,7 +443,6 @@ class TestStencil(TestStencilBase):
         self.check(test_impl_seq, test_impl, n)
 
     @skip_unsupported
-    @tag('important')
     def test_stencil_call_const(self):
         """Tests numba.stencil call that has an index that can be inferred as
         constant from a unary expr. Otherwise, this would raise an error since
@@ -497,21 +509,20 @@ class TestStencil(TestStencilBase):
         np.testing.assert_almost_equal(parfor_output4, expected, decimal=3)
 
         # check error in regular Python path
-        with self.assertRaises(ValueError) as e:
+        with self.assertRaises(NumbaValueError) as e:
             test_impl4(4)
 
         self.assertIn("stencil kernel index is not constant, "
                       "'neighborhood' option required", str(e.exception))
         # check error in njit path
         # TODO: ValueError should be thrown instead of LoweringError
-        with self.assertRaises(LoweringError) as e:
+        with self.assertRaises((LoweringError, NumbaValueError)) as e:
             njit(test_impl4)(4)
 
         self.assertIn("stencil kernel index is not constant, "
                       "'neighborhood' option required", str(e.exception))
 
     @skip_unsupported
-    @tag('important')
     def test_stencil_parallel_off(self):
         """Tests 1D numba.stencil calls without parallel translation
            turned off.
@@ -522,6 +533,113 @@ class TestStencil(TestStencilBase):
         cpfunc = self.compile_parallel(test_impl, (numba.float64[:],), stencil=False)
         self.assertNotIn('@do_scheduling', cpfunc.library.get_llvm_str())
 
+    @skip_unsupported
+    def test_stencil_nested1(self):
+        """Tests whether nested stencil decorator works.
+        """
+        @njit(parallel=True)
+        def test_impl(n):
+            @stencil
+            def fun(a):
+                c = 2
+                return a[-c+1]
+            B = fun(n)
+            return B
+
+        def test_impl_seq(n):
+            B = np.zeros(len(n), dtype=int)
+            for i in range(1, len(n)):
+                B[i] = n[i-1]
+            return B
+
+        n = np.arange(10)
+        np.testing.assert_equal(test_impl(n), test_impl_seq(n))
+
+    @skip_unsupported
+    def test_out_kwarg_w_cval(self):
+        """ Issue #3518, out kwarg did not work with cval."""
+        # test const value that matches the arg dtype, and one that can be cast
+        const_vals = [7, 7.0]
+
+        def kernel(a):
+            return (a[0, 0] - a[1, 0])
+
+        for const_val in const_vals:
+            stencil_fn = numba.stencil(kernel, cval=const_val)
+
+            def wrapped():
+                A = np.arange(12).reshape((3, 4))
+                ret = np.ones_like(A)
+                stencil_fn(A, out=ret)
+                return ret
+
+            # stencil function case
+            A = np.arange(12).reshape((3, 4))
+            expected = np.full_like(A, -4)
+            expected[-1, :] = const_val
+            ret = np.ones_like(A)
+            stencil_fn(A, out=ret)
+            np.testing.assert_almost_equal(ret, expected)
+
+            # wrapped function case, check njit, then njit(parallel=True)
+            impls = self.compile_all(wrapped,)
+            for impl in impls:
+                got = impl.entry_point()
+                np.testing.assert_almost_equal(got, expected)
+
+        # now check exceptions for cval dtype mismatch with out kwarg dtype
+        stencil_fn = numba.stencil(kernel, cval=1j)
+
+        def wrapped():
+            A = np.arange(12).reshape((3, 4))
+            ret = np.ones_like(A)
+            stencil_fn(A, out=ret)
+            return ret
+
+        A = np.arange(12).reshape((3, 4))
+        ret = np.ones_like(A)
+        with self.assertRaises(NumbaValueError) as e:
+            stencil_fn(A, out=ret)
+        msg = "cval type does not match stencil return type."
+        self.assertIn(msg, str(e.exception))
+
+        for compiler in [self.compile_njit, self.compile_parallel]:
+            try:
+                compiler(wrapped,())
+            except(NumbaValueError, LoweringError) as e:
+                self.assertIn(msg, str(e))
+            else:
+                raise AssertionError("Expected error was not raised")
+
+    @skip_unsupported
+    def test_out_kwarg_w_cval_np_attr(self):
+        """ Test issue #7286 where the cval is a np attr/string-based numerical
+        constant"""
+        for cval in (np.nan, np.inf, -np.inf, float('inf'), -float('inf')):
+            def kernel(a):
+                return (a[0, 0] - a[1, 0])
+
+            stencil_fn = numba.stencil(kernel, cval=cval)
+
+            def wrapped():
+                A = np.arange(12.).reshape((3, 4))
+                ret = np.ones_like(A)
+                stencil_fn(A, out=ret)
+                return ret
+
+            # stencil function case
+            A = np.arange(12.).reshape((3, 4))
+            expected = np.full_like(A, -4)
+            expected[-1, :] = cval
+            ret = np.ones_like(A)
+            stencil_fn(A, out=ret)
+            np.testing.assert_almost_equal(ret, expected)
+
+            # wrapped function case, check njit, then njit(parallel=True)
+            impls = self.compile_all(wrapped,)
+            for impl in impls:
+                got = impl.entry_point()
+                np.testing.assert_almost_equal(got, expected)
 
 
 class pyStencilGenerator:
@@ -659,6 +777,9 @@ class pyStencilGenerator:
             """
             Generates an ast.Num of value `value`
             """
+            # pretend bools are ints, ast has no boolean literal support
+            if isinstance(value, bool):
+                return ast.Num(int(value))
             if abs(value) >= 0:
                 return ast.Num(value)
             else:
@@ -755,59 +876,216 @@ class pyStencilGenerator:
             indexing.
             """
 
-            node = self.generic_visit(node)
-            if (node.value.id in self._argnames) and (
-                    node.value.id not in self._standard_indexing):
-                # 2D index
-                if isinstance(node.slice.value, ast.Tuple):
-                    idx = []
-                    for x, val in enumerate(node.slice.value.elts):
-                        useval = self._const_assigns.get(val, val)
-                        idx.append(
-                            ast.BinOp(
-                                left=ast.Name(
-                                    id=self._id_pat %
-                                    self.ids[x],
-                                    ctx=ast.Load()),
+            def handle2dindex(node):
+                idx = []
+                for x, val in enumerate(node.slice.value.elts):
+                    useval = self._const_assigns.get(val, val)
+                    idx.append(
+                        ast.BinOp(
+                            left=ast.Name(
+                                id=self._id_pat % self.ids[x],
+                                ctx=ast.Load()),
+                            op=ast.Add(),
+                            right=useval,
+                            ctx=ast.Load()))
+                if self._idx_len == -1:
+                    self._idx_len = len(idx)
+                else:
+                    if(self._idx_len != len(idx)):
+                        raise ValueError(
+                            "Relative indexing mismatch detected")
+                if isinstance(node.ctx, ast.Store):
+                    msg = ("Assignments to array passed to "
+                            "stencil kernels is not allowed")
+                    raise ValueError(msg)
+                context = ast.Load()
+                newnode = ast.Subscript(
+                    value=node.value,
+                    slice=ast.Index(
+                        value=ast.Tuple(
+                            elts=idx,
+                            ctx=ast.Load()),
+                        ctx=ast.Load()),
+                    ctx=context)
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+
+                # now work out max/min for index ranges i.e. stencil size
+                if self._mins is None and self._maxes is None:
+                    # first pass
+                    self._mins = [self._imax] * self._idx_len
+                    self._maxes = [self._imin] * self._idx_len
+
+                if not self._neighborhood:
+                    for x, lnode in enumerate(node.slice.value.elts):
+                        if isinstance(lnode, ast.Num) or\
+                                isinstance(lnode, ast.UnaryOp):
+                            relvalue = self.get_val_from_num(lnode)
+                        elif (hasattr(lnode, 'id') and
+                                lnode.id in self._const_assigns):
+                            relvalue = self._const_assigns[lnode.id]
+                        else:
+                            raise ValueError(
+                                "Cannot interpret indexing value")
+                        if relvalue < self._mins[x]:
+                            self._mins[x] = relvalue
+                        if relvalue > self._maxes[x]:
+                            self._maxes[x] = relvalue
+                else:
+                    for x, lnode in enumerate(self._neighborhood):
+                        self._mins[x] = self._neighborhood[x][0]
+                        self._maxes[x] = self._neighborhood[x][1]
+
+                return newnode
+
+            def handle1dindex(node):
+                useval = self._const_assigns.get(
+                    node.slice.value, node.slice.value)
+                idx = ast.BinOp(left=ast.Name(
+                                id=self._id_pat % self.ids[0],
+                                ctx=ast.Load()),
                                 op=ast.Add(),
                                 right=useval,
-                                ctx=ast.Load()))
-                    if self._idx_len == -1:
-                        self._idx_len = len(idx)
+                                ctx=ast.Load())
+                if self._idx_len == -1:
+                    self._idx_len = 1
+                else:
+                    if(self._idx_len != 1):
+                        raise ValueError(
+                            "Relative indexing mismatch detected")
+                if isinstance(node.ctx, ast.Store):
+                    msg = ("Assignments to array passed to "
+                            "stencil kernels is not allowed")
+                    raise ValueError(msg)
+                context = ast.Load()
+                newnode = ast.Subscript(
+                    value=node.value,
+                    slice=ast.Index(
+                        value=idx,
+                        ctx=ast.Load()),
+                    ctx=context)
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+
+                # now work out max/min for index ranges i.e. stencil size
+                if self._mins is None and self._maxes is None:
+                    # first pass
+                    self._mins = [self._imax, ]
+                    self._maxes = [self._imin, ]
+
+                if not self._neighborhood:
+                    if isinstance(node.slice.value, ast.Num) or\
+                            isinstance(node.slice.value, ast.UnaryOp):
+                        relvalue = self.get_val_from_num(node.slice.value)
+                    elif (hasattr(node.slice.value, 'id') and
+                            node.slice.value.id in self._const_assigns):
+                        relvalue = self._const_assigns[node.slice.value.id]
                     else:
-                        if(self._idx_len != len(idx)):
-                            raise ValueError(
-                                "Relative indexing mismatch detected")
-                    if isinstance(node.ctx, ast.Store):
-                        msg = ("Assignments to array passed to "
-                               "stencil kernels is not allowed")
-                        raise ValueError(msg)
-                    context = ast.Load()
-                    newnode = ast.Subscript(
-                        value=node.value,
-                        slice=ast.Index(
-                            value=ast.Tuple(
-                                elts=idx,
+                        raise ValueError("Cannot interpret indexing value")
+                    if relvalue < self._mins[0]:
+                        self._mins[0] = relvalue
+                    if relvalue > self._maxes[0]:
+                        self._maxes[0] = relvalue
+                else:
+                    self._mins[0] = self._neighborhood[0][0]
+                    self._maxes[0] = self._neighborhood[0][1]
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+                return newnode
+
+            def computeSlice(i, node):
+                def gen_idx(val, x):
+                    useval = self._const_assigns.get(val, val)
+                    value = self.get_val_from_num(val)
+                    tmp = ast.BinOp(
+                            left=ast.Name(
+                                id=self._id_pat % self.ids[x],
                                 ctx=ast.Load()),
-                            ctx=ast.Load()),
-                        ctx=context)
-                    ast.copy_location(newnode, node)
-                    ast.fix_missing_locations(newnode)
+                            op=ast.Add(),
+                            right=useval,
+                            ctx=ast.Load())
+                    ast.copy_location(tmp, node)
+                    ast.fix_missing_locations(tmp)
+                    return tmp
+                newnode = ast.Slice(gen_idx(node.lower, i),
+                                 gen_idx(node.upper, i),
+                                 node.step)
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+                return newnode
 
-                    # now work out max/min for index ranges i.e. stencil size
-                    if self._mins is None and self._maxes is None:
-                        # first pass
-                        self._mins = [self._imax] * self._idx_len
-                        self._maxes = [self._imin] * self._idx_len
+            def computeIndex(i, node):
+                useval = self._const_assigns.get(node.value, node.value)
+                idx = ast.BinOp(left=ast.Name(
+                                id=self._id_pat % self.ids[i],
+                                ctx=ast.Load()),
+                                op=ast.Add(),
+                                right=useval,
+                                ctx=ast.Load())
+                newnode = ast.Index(value=idx, ctx=ast.Load())
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+                return newnode
 
-                    if not self._neighborhood:
-                        for x, lnode in enumerate(node.slice.value.elts):
-                            if isinstance(lnode, ast.Num) or\
-                                    isinstance(lnode, ast.UnaryOp):
-                                relvalue = self.get_val_from_num(lnode)
-                            elif (hasattr(lnode, 'id') and
-                                  lnode.id in self._const_assigns):
-                                relvalue = self._const_assigns[lnode.id]
+            def handleExtSlice(node):
+                idx = []
+                for i, val in enumerate(node.slice.dims):
+                    if isinstance(val, ast.Slice):
+                        idx.append(computeSlice(i, val))
+                    if isinstance(val, ast.Index):
+                        idx.append(computeIndex(i, val))
+                    # TODO: handle more node types
+                if self._idx_len == -1:
+                    self._idx_len = len(node.slice.dims)
+                else:
+                    if(self._idx_len != len(node.slice.dims)):
+                        raise ValueError(
+                            "Relative indexing mismatch detected")
+
+                if isinstance(node.ctx, ast.Store):
+                    msg = ("Assignments to array passed to "
+                            "stencil kernels is not allowed")
+                    raise ValueError(msg)
+                context = ast.Load()
+                newnode = ast.Subscript(
+                    value=node.value,
+                    slice=ast.ExtSlice(
+                        dims=idx,
+                        ctx=ast.Load()),
+                        ctx=context
+                        )
+
+                # now work out max/min for index ranges i.e. stencil size
+                if self._mins is None and self._maxes is None:
+                    # first pass
+                    self._mins = [self._imax] * self._idx_len
+                    self._maxes = [self._imin] * self._idx_len
+
+                if not self._neighborhood:
+                    for x, anode in enumerate(node.slice.dims):
+                        if isinstance(anode, ast.Slice):
+                            for lnode in [anode.lower, anode.upper]:
+                                if isinstance(lnode, ast.Num) or\
+                                        isinstance(lnode, ast.UnaryOp):
+                                    relvalue = self.get_val_from_num(lnode)
+                                elif (hasattr(lnode, 'id') and
+                                        lnode.id in self._const_assigns):
+                                    relvalue = self._const_assigns[lnode.id]
+                                else:
+                                    raise ValueError(
+                                        "Cannot interpret indexing value")
+                                if relvalue < self._mins[x]:
+                                    self._mins[x] = relvalue
+                                if relvalue > self._maxes[x]:
+                                    self._maxes[x] = relvalue
+                        else:
+                            val = anode.value
+                            if isinstance(val, ast.Num) or\
+                                    isinstance(val, ast.UnaryOp):
+                                relvalue = self.get_val_from_num(val)
+                            elif (hasattr(val, 'id') and
+                                    val.id in self._const_assigns):
+                                relvalue = self._const_assigns[val.id]
                             else:
                                 raise ValueError(
                                     "Cannot interpret indexing value")
@@ -815,67 +1093,79 @@ class pyStencilGenerator:
                                 self._mins[x] = relvalue
                             if relvalue > self._maxes[x]:
                                 self._maxes[x] = relvalue
-                    else:
-                        for x, lnode in enumerate(self._neighborhood):
-                            self._mins[x] = self._neighborhood[x][0]
-                            self._maxes[x] = self._neighborhood[x][1]
+                else:
+                    for x, lnode in enumerate(self._neighborhood):
+                        self._mins[x] = self._neighborhood[x][0]
+                        self._maxes[x] = self._neighborhood[x][1]
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+                return newnode
 
-                    return newnode
+
+            def handleSlice(node):
+                idx = computeSlice(0, node.slice)
+                idx.ctx=ast.Load()
+                if isinstance(node.ctx, ast.Store):
+                    msg = ("Assignments to array passed to "
+                            "stencil kernels is not allowed")
+                    raise ValueError(msg)
+                context = ast.Load()
+                newnode = ast.Subscript(
+                    value=node.value,
+                    slice=idx,
+                    ctx=context)
+                ast.copy_location(newnode, node)
+                ast.fix_missing_locations(newnode)
+
+                if self._idx_len == -1:
+                    self._idx_len = 1
+                else:
+                    if(self._idx_len != 1):
+                        raise ValueError(
+                            "Relative indexing mismatch detected")
+
+                # now work out max/min for index ranges i.e. stencil size
+                if self._mins is None and self._maxes is None:
+                    # first pass
+                    self._mins = [self._imax]
+                    self._maxes = [self._imin]
+
+                if not self._neighborhood:
+                    if isinstance(node.slice.value, ast.Num) or\
+                            isinstance(node.slice.value, ast.UnaryOp):
+                        relvalue = self.get_val_from_num(node.slice.value)
+                    elif (hasattr(node.slice.value, 'id') and
+                            node.slice.value.id in self._const_assigns):
+                        relvalue = self._const_assigns[node.slice.value.id]
+                    else:
+                        raise ValueError("Cannot interpret indexing value")
+                    if relvalue < self._mins[0]:
+                        self._mins[0] = relvalue
+                    if relvalue > self._maxes[0]:
+                        self._maxes[0] = relvalue
+                else:
+                    self._mins[0] = self._neighborhood[0][0]
+                    self._maxes[0] = self._neighborhood[0][1]
+
+                return newnode
+
+
+            node = self.generic_visit(node)
+            if (node.value.id in self._argnames) and (
+                    node.value.id not in self._standard_indexing):
+
+                # fancy slice
+                if isinstance(node.slice, ast.ExtSlice):
+                    return handleExtSlice(node)
+                # plain slice
+                if isinstance(node.slice, ast.Slice):
+                    return handleSlice(node)
+                # 2D index
+                if isinstance(node.slice.value, ast.Tuple):
+                    return handle2dindex(node)
                 # 1D index
                 elif isinstance(node.slice, ast.Index):
-                    useval = self._const_assigns.get(
-                        node.slice.value, node.slice.value)
-                    idx = ast.BinOp(left=ast.Name(
-                                    id=self._id_pat %
-                                    self.ids[0],
-                                    ctx=ast.Load()),
-                                    op=ast.Add(),
-                                    right=useval,
-                                    ctx=ast.Load())
-                    if self._idx_len == -1:
-                        self._idx_len = 1
-                    else:
-                        if(self._idx_len != 1):
-                            raise ValueError(
-                                "Relative indexing mismatch detected")
-                    if isinstance(node.ctx, ast.Store):
-                        msg = ("Assignments to array passed to "
-                               "stencil kernels is not allowed")
-                        raise ValueError(msg)
-                    context = ast.Load()
-                    newnode = ast.Subscript(
-                        value=node.value,
-                        slice=ast.Index(
-                            value=idx,
-                            ctx=ast.Load()),
-                        ctx=context)
-                    ast.copy_location(newnode, node)
-                    ast.fix_missing_locations(newnode)
-
-                    # now work out max/min for index ranges i.e. stencil size
-                    if self._mins is None and self._maxes is None:
-                        # first pass
-                        self._mins = [self._imax, ]
-                        self._maxes = [self._imin, ]
-
-                    if not self._neighborhood:
-                        if isinstance(node.slice.value, ast.Num) or\
-                                isinstance(node.slice.value, ast.UnaryOp):
-                            relvalue = self.get_val_from_num(node.slice.value)
-                        elif (hasattr(node.slice.value, 'id') and
-                              node.slice.value.id in self._const_assigns):
-                            relvalue = self._const_assigns[node.slice.value.id]
-                        else:
-                            raise ValueError("Cannot interpret indexing value")
-                        if relvalue < self._mins[0]:
-                            self._mins[0] = relvalue
-                        if relvalue > self._maxes[0]:
-                            self._maxes[0] = relvalue
-                    else:
-                        self._mins[0] = self._neighborhood[0][0]
-                        self._maxes[0] = self._neighborhood[0][1]
-
-                    return newnode
+                    return handle1dindex(node)
                 else:  # unknown
                     raise ValueError("Unhandled subscript")
             else:
@@ -1028,11 +1318,8 @@ class pyStencilGenerator:
             returner = self.gen_return(retvar)
             ast.copy_location(returner, node)
 
-            if _py27:
-                add_kwarg = [ast.Name('neighborhood', ast.Param())]
-            else:
-                add_kwarg = [ast.arg('neighborhood', None)]
-            defaults = [ast.Name(id='None', ctx=ast.Load())]
+            add_kwarg = [ast.arg('neighborhood', None)]
+            defaults = []
 
             newargs = ast.arguments(
                 args=node.args.args +
@@ -1041,7 +1328,8 @@ class pyStencilGenerator:
                 vararg=None,
                 kwarg=None,
                 kwonlyargs=[],
-                kw_defaults=[])
+                kw_defaults=[],
+                posonlyargs=[])
             new = ast.FunctionDef(
                 name='__%s' %
                 node.name,
@@ -1076,11 +1364,7 @@ class pyStencilGenerator:
             if self._argnames is not None or self._kwargnames is not None:
                 raise RuntimeError("multiple definition of function/args?")
 
-            if _py27:
-                attr = 'id'
-            else:
-                attr = 'arg'
-
+            attr = 'arg'
             self._argnames = [getattr(x, attr) for x in node.args.args]
             if node.args.kwarg:
                 self._kwargnames = [x.arg for x in node.args.kwarg]
@@ -1332,6 +1616,7 @@ class TestManyStencils(TestStencilBase):
         stencil_args.update(options)
 
         expected_present = True
+
         try:
             # ast impl
             ast_impl = pyStencil(func_or_mode=pyfunc, **options)
@@ -1683,7 +1968,7 @@ class TestManyStencils(TestStencilBase):
 
         def kernel(a):
             return 1.
-        self.check(kernel, a, expected_exception=[ValueError, LoweringError])
+        self.check(kernel, a, expected_exception=[ValueError, NumbaValueError])
 
     def test_basic26(self):
         """3d arr"""
@@ -1715,7 +2000,8 @@ class TestManyStencils(TestStencilBase):
 
         def kernel(a):
             return a[0, int(np.cos(0))]
-        self.check(kernel, a, expected_exception=[ValueError, LoweringError])
+        self.check(kernel, a, expected_exception=[ValueError, NumbaValueError,
+                                                  LoweringError])
 
     def test_basic30(self):
         """signed zeros"""
@@ -1760,7 +2046,8 @@ class TestManyStencils(TestStencilBase):
 
         def kernel(a):
             return a[np.int8(1), 0]
-        self.check(kernel, a, expected_exception=[ValueError, LoweringError])
+        self.check(kernel, a, expected_exception=[ValueError, NumbaValueError,
+                                                  LoweringError])
 
     def test_basic33(self):
         """add 0d array"""
@@ -1784,9 +2071,9 @@ class TestManyStencils(TestStencilBase):
             return a[0, 1]
         a = np.arange(12.).reshape(3, 4)
         ex = self.exception_dict(
-            stencil=ValueError,
+            stencil=NumbaValueError,
             parfor=ValueError,
-            njit=LoweringError)
+            njit=NumbaValueError)
         self.check(kernel, a, options={'cval': 5}, expected_exception=ex)
 
     def test_basic36(self):
@@ -1809,9 +2096,9 @@ class TestManyStencils(TestStencilBase):
             return a[0, 1] + a[0, -1] + a[1, -1] + a[1, -1]
         a = np.arange(12.).reshape(3, 4)
         ex = self.exception_dict(
-            stencil=ValueError,
+            stencil=NumbaValueError,
             parfor=ValueError,
-            njit=LoweringError)
+            njit=NumbaValueError)
         self.check(kernel, a, options={'cval': 1.j}, expected_exception=ex)
 
     def test_basic39(self):
@@ -1943,7 +2230,7 @@ class TestManyStencils(TestStencilBase):
             kernel, a, b, options={
                 'standard_indexing': [
                     'a', 'b']}, expected_exception=[
-                ValueError, LoweringError])
+                ValueError, NumbaValueError])
 
     def test_basic52(self):
         """3 args, standard_indexing on middle arg """
@@ -2084,9 +2371,9 @@ class TestManyStencils(TestStencilBase):
         b = np.arange(12).reshape(3, 4)
         ex = self.exception_dict(
             pyStencil=ValueError,
-            stencil=ValueError,
+            stencil=NumbaValueError,
             parfor=ValueError,
-            njit=LoweringError)
+            njit=NumbaValueError)
         self.check(
             kernel,
             a,
@@ -2108,7 +2395,7 @@ class TestManyStencils(TestStencilBase):
                 'standard_indexing': 'a'},
             expected_exception=[
                 ValueError,
-                LoweringError])
+                NumbaValueError])
 
     def test_basic65(self):
         """basic induced neighborhood test"""
@@ -2437,9 +2724,9 @@ class TestManyStencils(TestStencilBase):
         a = np.arange(12.).reshape(3, 4)
         ex = self.exception_dict(
             pyStencil=ValueError,
-            stencil=ValueError,
+            stencil=NumbaValueError,
             parfor=ValueError,
-            njit=LoweringError)
+            njit=NumbaValueError)
         self.check(
             kernel,
             a,
@@ -2479,6 +2766,77 @@ class TestManyStencils(TestStencilBase):
                 'neighborhood': (
                     (-3, 0), (-3, 0)), 'standard_indexing': 'b', 'cval': 1.5})
 
+    def test_basic91(self):
+        """ Issue #3454, const(int) == const(int) evaluating incorrectly. """
+        def kernel(a):
+            b = 0
+            if(2 == 0):
+                b = 2
+            return a[0, 0] + b
+
+        a = np.arange(10. * 20.).reshape(10, 20)
+        self.check(kernel, a)
+
+    def test_basic92(self):
+        """ Issue #3497, bool return type evaluating incorrectly. """
+        def kernel(a):
+            return (a[-1, -1] ^ a[-1, 0] ^ a[-1, 1] ^
+                    a[0, -1]  ^ a[0, 0]  ^ a[0, 1]  ^
+                    a[1, -1]  ^ a[1, 0]  ^ a[1, 1])
+
+        A = np.array(np.arange(20) % 2).reshape(4, 5).astype(np.bool_)
+        self.check(kernel, A)
+
+    def test_basic93(self):
+        """ Issue #3497, bool return type evaluating incorrectly. """
+        def kernel(a):
+            return (a[-1, -1] ^ a[-1, 0] ^ a[-1, 1] ^
+                    a[0, -1]  ^ a[0, 0]  ^ a[0, 1]  ^
+                    a[1, -1]  ^ a[1, 0]  ^ a[1, 1])
+
+        A = np.array(np.arange(20) % 2).reshape(4, 5).astype(np.bool_)
+        self.check(kernel, A, options={'cval': True})
+
+    def test_basic94(self):
+        """ Issue #3528. Support for slices. """
+        def kernel(a):
+            return np.median(a[-1:2, -1:2])
+        a = np.arange(20, dtype=np.uint32).reshape(4, 5)
+        self.check(kernel, a, options={'neighborhood': ((-1, 1), (-1, 1),)})
+
+    @unittest.skip("not yet supported")
+    def test_basic95(self):
+        """ Slice, calculate neighborhood. """
+        def kernel(a):
+            return np.median(a[-1:2, -3:4])
+        a = np.arange(20, dtype=np.uint32).reshape(4, 5)
+        self.check(kernel, a)
+
+    def test_basic96(self):
+        """ 1D slice. """
+        def kernel(a):
+            return np.median(a[-1:2])
+        a = np.arange(20, dtype=np.uint32)
+        self.check(kernel, a, options={'neighborhood': ((-1, 1),)})
+
+    @unittest.skip("not yet supported")
+    def test_basic97(self):
+        """ 2D slice and index. """
+        def kernel(a):
+            return np.median(a[-1:2, 3])
+        a = np.arange(20, dtype=np.uint32).reshape(4, 5)
+        self.check(kernel, a)
+
+    def test_basic98(self):
+        """ Test issue #7286 where the cval is a np attr/string-based numerical
+        constant"""
+        for cval in (np.nan, np.inf, -np.inf, float('inf'), -float('inf')):
+            def kernel(a):
+                return a[0, 0]
+            a = np.arange(6.).reshape((2, 3))
+            self.check(kernel, a, options={'neighborhood': ((-1, 1), (-1, 1),),
+                       'cval':cval})
+
 
 if __name__ == "__main__":
     unittest.main()
